#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Untitled Messenger Bag Display Digital Art
# Copyright 2019 Ben Lincoln
# https://www.beneaththewaves.net/
#

# This file is part of Untitled Messenger Bag Display Digital Art.

# Untitled Messenger Bag Display Digital Art is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU General Public License as published by
# the Free Software Foundation.

# Untitled Messenger Bag Display Digital Art is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Untitled Messenger Bag Display Digital Art (in the file LICENSE.txt).  
# If not, see <http://www.gnu.org/licenses/>.

import datetime
from math import *
import inspect
import math
import os
import pygame
import random
import sys
import time

class on_screen_object(object):
	def __init__(self):
		is_alive = True
		self.position = (0, 0)
		self.last_position = (self.position[0], self.position[1])
		self.rotation_degrees = 0.0
		self.last_rotation_degrees = self.rotation_degrees
		self.velocity = (1.0, 1.0)
		self.angular_velocity = 1.0
		self.size = (40, 40)
		self.set_size((40, 40))
		self.colour = (230, 255, 245)
		self.alpha = 128
		self.chromakey_colour = (0, 0, 0)
		# Enable blit_instead_of_draw for per-object alpha blending
		# VERY slow with large numbers of objects
		self.blit_instead_of_draw = False
		self.blit_buffer = pygame.Surface((screen_width, screen_height))
		self.cached_points = []
		self.cached_points_rotated = []
		self.velocity_change_percent_change = 0
		# +/- 5
		self.velocity_change_max = 2
		self.velocity_max = 11
		self.bounce_at_edge = True
		self.fill_polygon = False
		#self.fill_polygon = True
		self.line_width = 0
		
	def set_size(self, new_size):
		self.last_size = (self.size[0], self.size[1])
		self.size = (new_size[0], new_size[1])

	def update_velocity(self):
		if random.randint(0, 100) < self.velocity_change_percent_change:
			velocity_change = random.randint(0, (self.velocity_change_max * 2)) - self.velocity_change_max
			new_velocity = float(self.velocity[0] + velocity_change)
			if new_velocity > self.velocity_max:
				new_velocity = float(self.velocity_max)
			if new_velocity < (self.velocity_max * -1):
				new_velocity = float(self.velocity_max * -1)
			self.velocity = (new_velocity, self.velocity[1])
		if random.randint(0, 100) < self.velocity_change_percent_change:
			velocity_change = random.randint(0, (self.velocity_change_max * 2)) - self.velocity_change_max
			new_velocity = float(self.velocity[1] + velocity_change)
			if new_velocity > self.velocity_max:
				new_velocity = float(self.velocity_max)
			if new_velocity < (self.velocity_max * -1):
				new_velocity = float(self.velocity_max * -1)
			self.velocity = float(self.velocity[0], new_velocity)
			
		# bounce
		if self.bounce_at_edge:
			bounce_x = False
			if self.position[0] < 0 and self.velocity[0] < 0:
				bounce_x = True
			if self.position[0] > screen_width and self.velocity[0] > 0:
				bounce_x = True
			if bounce_x:
				self.velocity = (self.velocity[0] * -1, self.velocity[1])

			bounce_y = False
			if self.position[1] < 0 and self.velocity[1] < 0:
				bounce_y = True
			if self.position[1] > screen_height and self.velocity[1] > 0:
				bounce_y = True
			if bounce_y:
				self.velocity = (self.velocity[0], self.velocity[1] * -1)
			
	def update_position(self):
		self.last_position = (self.position[0], self.position[1])
		self.last_rotation_degrees = self.rotation_degrees
		self.position = (self.position[0] + self.velocity[0], self.position[1] + self.velocity[1])
		self.rotation_degrees += self.angular_velocity
		while self.rotation_degrees > 359:
			self.rotation_degrees -= 360.0
		while self.rotation_degrees < -359:
			self.rotation_degrees += 360.0
		#print("Debug: rotation = %s" % (self.rotation_degrees))
		#return position
		
	def get_points(self):
		needs_update_base = False
		needs_update_rotation = False
		if self.position[0] != self.last_position[0] or self.position[1] != self.last_position[1]:
			needs_update_base = True
			#print("Position has changed")
		if len(self.cached_points) < 1:
			needs_update_base = True
		if self.size[0] != self.last_size[0] or self.size[1] != self.last_size[1]:
			needs_update_base = True
			#print("Size has changed")
		self.last_size = (self.size[0], self.size[1])
		if self.angular_velocity != 0:
			needs_update_rotation = True
			#print("Object is rotating")
		if self.last_rotation_degrees != self.rotation_degrees:
			needs_update_rotation = True
			#print("Object has been rotated")
		points = []
		if needs_update_base:
			needs_update_rotation = True
			points = self.get_points_inner()
			for p in range(0, len(points)):
				self.cached_points.append(points[p])
		else:
			#for p in self.cached_points:
			#	points.append(p)
			for p in range(0, len(self.cached_points)):
				points.append(self.cached_points[p])
		if needs_update_rotation:
			points_temp = on_screen_object.rotate_points(self.position, points, self.rotation_degrees)
			points = []
			#for p in points_temp:
			#	points.append(p)
			#	self.cached_points_rotated.append(p)
			for p in range(0, len(points_temp)):
				points.append(points_temp[p])
				self.cached_points_rotated.append(points_temp[p])
		else:
			#for p in self.cached_points_rotated:
			#	points.append(p)
			points = []
			for p in range(0, len(self.cached_points_rotated)):
				points.append(self.cached_points_rotated[p])
		return points

	def get_points_inner(self):
		points = []
		base_x = self.position[0]
		base_y = self.position[1]
		width = self.size[0]
		height = self.size[1]
		points.append(self.position)
		points.append((base_x, base_y + height))
		points.append((base_x + width, base_y + height))
		points.append((base_x + width, base_y))
		points.append(self.position)
		return points
		
	def update(self):
		self.update_position()
		self.update_velocity()

	@staticmethod
	def rotate(origin, point, angle):
		"""
		Rotate a point counterclockwise by a given angle around a given origin.

		The angle should be given in radians.
		https://stackoverflow.com/questions/34372480/rotate-point-about-another-point-in-degrees-python
		"""
		ox, oy = origin
		px, py = point

		qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
		qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
		return qx, qy
	
	@staticmethod
	def rotate_points(origin, points, degrees):
		if degrees != 0:
			result = []
			for p in points:
				result.append(on_screen_object.rotate(origin, p, math.radians(degrees)))
			return result
		else:
			return points
	
	def draw(self, surf):
		points = self.get_points()
		if self.blit_instead_of_draw:
			self.blit_buffer.fill(self.chromakey_colour)
			self.blit_buffer.set_alpha(self.alpha)
			self.blit_buffer.set_colorkey(self.chromakey_colour)
			if self.fill_polygon and len(points) > 2:
				#print("Debug: filling polygon")
				#pygame.draw.polygon(self.blit_buffer, self.colour, points[1:])
				#points.reverse()
				pygame.draw.polygon(self.blit_buffer, self.colour, points)
			else:
				if self.line_width == 0:
					pygame.draw.lines(self.blit_buffer, self.colour, False, points, 1)				
				#else:
				#	pygame.draw.polygon(self.blit_buffer, self.colour, points, width = self.line_width)
			surf.blit(self.blit_buffer, (0,0))
		else:
			if self.fill_polygon and len(points) > 2:
				#print("Debug: filling polygon")
				#points.reverse()
				pygame.draw.polygon(surf, self.colour, points)
				#pygame.draw.polygon(surf, (255, 255, 255), points)
			else:
				if self.line_width == 0:
					pygame.draw.lines(surf, self.colour, False, points, 1)
				#else:
				#	pygame.draw.polygon(surf, self.colour, points, width = self.line_width)
			
			
	@staticmethod
	def get_ngon_points(num_sides, radius, position):
		pi2 = 2 * 3.141593
		result = []

		for i in range(0, num_sides):
			result.append((round(cos(float(i) / float(num_sides) * pi2) * float(radius) + float(position[0])), round(sin(float(i) / float(num_sides) * pi2) * float(radius) + float(position[1]))))

		result.append(result[0])
			
		return result
		
	@staticmethod
	def get_factors(integer):
		result = [integer]
		# there is probably a more efficient way to do this, but this is fine for the relatively small numbers used in this program
		current_value = float(math.floor(float(integer) / 2.0))
		base_value = float(integer)
		while current_value > 1.0:
			factor_test = base_value / current_value
			if factor_test % 1.0 == 0:
				result.append(int(factor_test))
			current_value -= 1
		result.append(1)
		print("Factors of %s: %s" % (integer, result))
		return result

class arbitrary_geometry(on_screen_object):
	def __init__(self):
		super(arbitrary_geometry, self).__init__()
		self.points = []

	def get_points_inner(self):
		points = []
		base_x = self.position[0]
		base_y = self.position[1]
		for p in self.points:
			points.append((p[0] + base_x, p[1] + base_y))
		return points
		
#class hexagon(on_screen_object):
#	def __init__(self):
#		super(hexagon, self).__init__()
#		self.num_sides = 6
#
#	def get_points_inner(self):
#		return self.get_ngon_points(6, self.size[0], self.position)
		
class polygon(on_screen_object):
	def __init__(self, num_sides):
		super(polygon, self).__init__()
		self.num_sides = num_sides

	def get_points_inner(self):
		return self.get_ngon_points(self.num_sides, self.size[0], self.position)
		
class polygram(polygon):
	def __init__(self, num_sides):
		super(polygon, self).__init__()
		self.num_sides = num_sides
		self.last_num_sides = num_sides
		self.point_step = polygram.get_point_step(num_sides)
		self.mandala_mode = False

	# How many points to skip when connecting points to make a star
	# This is surprisingly complicated if you've never done it before
	@staticmethod
	def get_point_step(num_sides):
		# the naive, hardcoded approach, which more or less works OK for about 20 or fewer points
		# n = 2
		# if num_sides > 6:
			# n = 3
		# if num_sides >= 9:
			# n = 4
		# if num_sides >= 11:
			# n = 5
		# if num_sides >= 13:
			# n = 6
		# if num_sides >= 13:
			# n = 6
			
		# # special cases
		# if num_sides == 14:
			# n = 5
			
		# the correct method for getting the pointiest possible stars that are drawn "like a pentagram", 
		# that is, by connecting all points in a single sequence instead of e.g. by overlaying multiple rotated copies of a polygon ("like a Star of David"),
		# except where this is impossible (can't draw a radially-symmetric hexagram except as a Star of David).
		if num_sides == 6:
			print("There is no way to draw a two-dimensional, radially-symmetric, unicursal 6-pointed star - using step value of 2")
			return 2
			
		factors_num_sides = on_screen_object.get_factors(num_sides)
		current_step = int(math.floor(float(num_sides) / 2.0))
		found_value = False
		while not found_value:
			factors_current_step = on_screen_object.get_factors(current_step)
			common_factors = []
			for f in factors_current_step:
				if f in factors_num_sides:
					common_factors.append(f)
			print("Common factors of %s and %s: %s" % (num_sides, current_step, common_factors))
			if len(common_factors) < 2:
				print("Found pointiest co-prime step value for a radially-symmetric, unicursal %s-pointed star: %s" % (num_sides, current_step))
				n = current_step
				found_value = True
			current_step -= 1
			if current_step < 1:
				print("Error: did not find any co-prime values for a radially-symmetric, unicursal %s-pointed star" % (num_sides))
				# fall back to the hardcoded method because the show must go on
				n = 2
				if num_sides > 6:
					n = 3
				if num_sides >= 9:
					n = 4
				if num_sides >= 11:
					n = 5
				if num_sides >= 13:
					n = 6
				if num_sides >= 13:
					n = 6					
				# special cases
				if num_sides == 14:
					n = 5
				found_value = True
			
		return n
	
#	def draw_inner(self, surf, colour, points):
#		for i in range(0, len(points)):
#			line_points = []
#			line_points.append(points[i])
#			second_num = (i + self.point_step) % self.num_sides
#			line_points.append(points[second_num])
#			#pygame.draw.lines(surf, colour, False, line_points, 1)
#			self.fill_polygon = True
#			if self.fill_polygon and len(line_points) > 2:
#				pygame.draw.polygon(surf, colour, line_points)
#			else:
#				if self.line_width == 0:
#					pygame.draw.lines(surf, colour, False, line_points, 1)		
#				#else:
#				#	pygame.draw.polygon(self.blit_buffer, self.colour, points, width = self.line_width)
				
	def draw_inner(self, surf, colour, points):
		line_points = []
		if self.mandala_mode:
			for i in range(0, len(points)):
				line_points.append(points[i])
				second_num = (i + self.point_step) % self.num_sides
				line_points.append(points[second_num])
		else:
			line_points.append(points[0])
			point_num = 0
			for i in range(1, len(points)):
				point_num = (point_num + self.point_step) % self.num_sides
				line_points.append(points[point_num])

		#pygame.draw.lines(surf, colour, False, line_points, 1)
		if self.fill_polygon and len(line_points) > 2:
			pygame.draw.polygon(surf, colour, line_points)
		else:
			if self.line_width == 0:
				pygame.draw.lines(surf, colour, False, line_points, 1)		
			#else:
			#	pygame.draw.polygon(self.blit_buffer, self.colour, points, width = self.line_width)
			
	def draw(self, surf):
		# in case the number of sides has changed
		if self.last_num_sides != self.num_sides:
			self.point_step = polygram.get_point_step(self.num_sides)
			self.last_num_sides = self.num_sides
		points = self.get_points()
		if self.blit_instead_of_draw:
			self.blit_buffer.fill(self.chromakey_colour)
			self.blit_buffer.set_alpha(self.alpha)
			self.blit_buffer.set_colorkey(self.chromakey_colour)
			self.draw_inner(self.blit_buffer, self.colour, points)
			surf.blit(self.blit_buffer, (0,0))
		else:
			self.draw_inner(surf, self.colour, points)

class ring_segment(object):
	def __init__(self):
		self.start_degree = 0.0
		self.end_degree = 359.0

class segmented_ring(on_screen_object):
	def __init__(self, outer_radius, thickness):
		super(segmented_ring, self).__init__()
		self.outer_radius = outer_radius
		self.thickness = thickness
		self.segments = []
		# for debugging
		rs_test1 = ring_segment()
		rs_test1.start_degree = 0.0
		rs_test1.end_degree = 45.0
		rs_test2 = ring_segment()
		rs_test2.start_degree = 90.0
		rs_test2.end_degree = 120.0
		rs_test3 = ring_segment()
		rs_test3.start_degree = 200.0
		rs_test3.end_degree = 240.0
		self.segments.append(rs_test1)
		self.segments.append(rs_test2)
		self.segments.append(rs_test3)
		self.colour = (255, 255, 255)

	def draw(self, surf):
		if self.blit_instead_of_draw:
			self.blit_buffer.fill(self.chromakey_colour)
			self.blit_buffer.set_alpha(self.alpha)
			self.blit_buffer.set_colorkey(self.chromakey_colour)
			self.draw_inner(self.blit_buffer, self.colour)
			surf.blit(self.blit_buffer, (0,0))
		else:
			self.draw_inner(surf, self.colour)
	
	def draw_inner(self, surf, colour):
		#surf_width = surf.get_width()
		#surf_height = surf.get_height()
		#shortest_edge = surf_width
		#length_diff = (surf_height - shortest_edge) / 2
		#draw_rect = ((0, length_diff), (shortest_edge, shortest_edge))
		#draw_window_start_x = 0
		#draw_window_start_y = length_diff
		#draw_window_end_x = shortest_edge
		#draw_window_end_y = shortest_edge
		#if surf_height < surf_width:
		#	shortest_edge = surf_height
		#	length_diff = (surf_width - shortest_edge) / 2
			#draw_rect = ((length_diff, 0), (shortest_edge, shortest_edge))
		#	draw_window_start_x = length_diff
		#	draw_window_start_y = 0
			#draw_rect = ((0, 0), (shortest_edge, shortest_edge))
			#print("Debug: surf_height < surf_width")
		#draw_rect = ((0, 0), (surf_width, surf_height))
		center_x_outer = self.position[0] - self.outer_radius
		center_y_outer = self.position[1] - self.outer_radius
		inner_radius = self.outer_radius - self.thickness
		center_x_inner = self.position[0] - inner_radius
		center_y_inner = self.position[1] - inner_radius
		#base_x = center_x - self.outer_radius
		#base_y = center_y - self.outer_radius
		#print("Debug: self.position = %s, center_x = %s, center_y = %s" % (self.position, center_x, center_y))
		#draw_rect_outer = ((center_x - self.outer_radius, center_y - self.outer_radius), (center_x + self.outer_radius, center_y + self.outer_radius))
		#draw_rect_outer = ((int(base_x), int(base_y)), (int(base_x + (self.outer_radius * 2.0)), int(base_y + (self.outer_radius * 2.0))))
		draw_rect_outer = ((center_x_outer, center_y_outer), (self.outer_radius * 2.0, self.outer_radius * 2.0))
		draw_rect_inner = ((center_x_inner, center_y_inner), (inner_radius * 2.0, inner_radius * 2.0))
		#print("Debug: draw_rect_outer = %s, %s, %s, %s" % (draw_rect_outer[0][0], draw_rect_outer[0][1], draw_rect_outer[1][0], draw_rect_outer[1][1]))
		#draw_rect_inner = ((center_x - (self.outer_radius - self.thickness), center_y - (self.outer_radius - self.thickness)), (center_x + (self.outer_radius - self.thickness), center_y + (self.outer_radius - self.thickness)))
		for i in range(0, len(self.segments)):
			start_x = self.position[0] + self.outer_radius
			end_x = self.position[0] + inner_radius
			both_y = self.position[1]

			start_angle = math.radians(self.segments[i].start_degree + self.rotation_degrees)
			end_angle = math.radians(self.segments[i].end_degree + self.rotation_degrees)
			
			line_1 = (self.rotate(self.position, (start_x, self.position[1]), start_angle * -1.0), self.rotate(self.position, (end_x, self.position[1]), start_angle * -1.0))
			line_2 = (self.rotate(self.position, (start_x, self.position[1]), end_angle * -1.0), self.rotate(self.position, (end_x, self.position[1]), end_angle * -1.0))
			
			#pygame.draw.arc(surf, colour, draw_rect, start_angle, end_angle, width = self.thickness)
			try:
				pygame.draw.arc(surf, colour, draw_rect_outer, start_angle, end_angle)
			except ValueError as ve:
				print("Error: PC load letter, WTF does that mean, or rather, 'width greater than ellipse radius', with values of draw_rect_outer[0][0] = %s, draw_rect_outer[0][1] = %s, draw_rect_outer[1][0] = %s, draw_rect_outer[1][1] = %s, start_angle = %s, end_angle = %s" % (draw_rect_outer[0][0], draw_rect_outer[0][1], draw_rect_outer[1][0], draw_rect_outer[1][1], start_angle, end_angle))
			try:
				pygame.draw.arc(surf, colour, draw_rect_inner, start_angle, end_angle)
			except ValueError as ve:
				print("Error: PC load letter, WTF does that mean, or rather, 'width greater than ellipse radius', with values of draw_rect_outer[0][0] = %s, draw_rect_outer[0][1] = %s, draw_rect_outer[1][0] = %s, draw_rect_outer[1][1] = %s, start_angle = %s, end_angle = %s" % (draw_rect_outer[0][0], draw_rect_outer[0][1], draw_rect_outer[1][0], draw_rect_outer[1][1], start_angle, end_angle))
			pygame.draw.line(surf, colour, line_1[0], line_1[1], 2)
			pygame.draw.line(surf, colour, line_2[0], line_2[1], 2)
			#pygame.draw.rect(surf, colour, draw_rect_outer, 1)

	def get_points(self):
		needs_update_base = False
		needs_update_rotation = False
		if self.position[0] != self.last_position[0] or self.position[1] != self.last_position[1]:
			needs_update_base = True
			#print("Position has changed")
		if len(self.cached_points) < 1:
			needs_update_base = True
		if self.size[0] != self.last_size[0] or self.size[1] != self.last_size[1]:
			needs_update_base = True
			#print("Size has changed")
		self.last_size = (self.size[0], self.size[1])
		if self.angular_velocity != 0:
			needs_update_rotation = True
			#print("Object is rotating")
		if self.last_rotation_degrees != self.rotation_degrees:
			needs_update_rotation = True
			#print("Object has been rotated")
		points = []
		if needs_update_base:
			needs_update_rotation = True
			points = self.get_points_inner()
			for p in points:
				self.cached_points.append(p)		
		else:
			for p in self.cached_points:
				points.append(p)
		if needs_update_rotation:
			points_temp = on_screen_object.rotate_points(self.position, points, self.rotation_degrees)
			points = []
			for p in points_temp:
				points.append(p)
				self.cached_points_rotated.append(p)
		else:
			for p in self.cached_points_rotated:
				points.append(p)
		return points

class on_screen_bitmap(on_screen_object):
	def __init__(self):
		super(on_screen_bitmap, self).__init__()
		self.velocity = (0, 0)
		self.angular_velocity = 0.0
		self.bitmap_file_path = ''
		self.bitmap = None
		self.bitmap_rect = None
		self.bitmap_needs_refresh = False
		self.cached_rotated_bitmap = None
		self.cached_rotated_bitmap_rect = None
		#self.blit_instead_of_draw = True
		
	def set_bitmap(self, file_path):
		self.bitmap_file_path = file_path
		self.bitmap = pygame.image.load(file_path)
		self.bitmap_rect = self.bitmap.get_rect()
		self.bitmap_needs_refresh = True
	
	def draw(self, surf):
		if self.angular_velocity != 0:
			needs_update_rotation = True
			#print("Object is rotating")
		if self.last_rotation_degrees != self.rotation_degrees:
			needs_update_rotation = True
			#print("Object has been rotated")
		if self.bitmap_needs_refresh:
			needs_update_rotation = True

		if needs_update_rotation:
			if self.rotation_degrees != 0.0:
				self.cached_rotated_bitmap = pygame.transform.rotate(self.bitmap, -1.0 * self.rotation_degrees)
			else:
				self.cached_rotated_bitmap = self.bitmap
			self.cached_rotated_bitmap_rect = self.cached_rotated_bitmap.get_rect()
		self.cached_rotated_bitmap_rect.center = self.position
		#screen.blit(myimage, imagerect)
		if self.blit_instead_of_draw:
			self.blit_buffer.fill(self.chromakey_colour)
			self.blit_buffer.set_alpha(self.alpha)
			self.blit_buffer.set_colorkey(self.chromakey_colour)
			self.blit_buffer.blit(self.cached_rotated_bitmap, self.cached_rotated_bitmap_rect)
			surf.blit(self.blit_buffer, (0,0))
		else:
			surf.blit(self.cached_rotated_bitmap, self.cached_rotated_bitmap_rect)
			#surf.blit(self.bitmap, self.bitmap_rect)
			#surf.blit(self.bitmap, (0,0))
			

class object_manager(object):
	def __init__(self, screen, screen_width, screen_height):
		self.num_boxes = 17
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
		self.rotation_degrees = 0.0
		self.last_surface = None
		#self.last_surface_scaling_factor = 16
		self.last_surface_scaling_factor = 32
		self.unchanging_surface_count = 0
		self.max_unchanged_frames = 3
		self.reset_on_no_drawing = True
		self.reset_on_no_objects = True
		self.add_post_blur = True
		self.current_settings_start_time = time.time()
		self.autoreset_interval_seconds = 15.6685
		self.display_seconds = 31.337
		#self.autoreset_interval_seconds = 3.137
	
		#self.box_min_size = 500
		#self.box_max_size = 5000
		self.box_min_size = 50
		self.box_max_size = 500
		
		self.frame_num = 0
		self.max_frame = 500
		
		#self.colour_background = (255, 255, 255)
		#self.colour_foreground = (0, 0, 0)

		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		self.on_screen_objects = []
		self.enable_chromakey = True
		self.alpha = -1
		
	def reset(self):		
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		for i in range(0, self.num_boxes):
			#new_obj = on_screen_object()
			new_obj = polygon(6)
			box_size = random.randint(self.box_min_size, self.box_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (self.screen_width / 2, self.screen_height / 2)
			new_obj.angular_velocity = random.randint(-10, 10)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = 1
			#initial_y_velocity = 1
			initial_y_velocity = 0
			if i % 2 == 0:
				initial_x_velocity = 0
			if i % 3 == 0:
				initial_y_velocity = 0
			if i % 4 == 0:
				initial_x_velocity = -1
				initial_y_velocity = 0
			if i % 5 == 0:
				initial_x_velocity = 0
				initial_y_velocity = -1
			if i % 6 == 0:
				initial_x_velocity = 0
				initial_y_velocity = 0
			#if i % 8 == 0:
			#	initial_x_velocity = -1
			#	initial_y_velocity = -1
			if i % 8 == 0:
				initial_x_velocity = -1
				initial_y_velocity = 0
			
			initial_x_velocity *= random.randint(1, 10)
			initial_y_velocity *= random.randint(1, 10)
				
			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
		
	def main_loop(self):
		if self.autoreset_interval_seconds > 0.0:
			elapsed = time.time() - self.current_settings_start_time
			if elapsed >= self.autoreset_interval_seconds:
				self.current_settings_start_time = time.time()
				self.reset()
		for i in range(0, len(self.on_screen_objects)):
			self.on_screen_objects[i].update()
			self.frame_num += 1
			if self.frame_num > self.max_frame:
				self.frame_num = 0

	def draw(self, surf):
		surf_temp = None
		if self.rotation_degrees != 0.0:
			surf_temp = pygame.transform.rotate(surf, self.rotation_degrees)
		else:
			surf_temp = surf.copy()

		num_objects = len(self.on_screen_objects)
		#print("Debug: %s on-screen objects" % (num_objects))
		for i in range(0, num_objects):
			#print("Debug: drawing object %s" % (i))
			if self.alpha != -1:
				#surf_temp.set_alpha(self.alpha)
				self.on_screen_objects[i].blit_instead_of_draw = True
				self.on_screen_objects[i].alpha = self.alpha
			else:
				self.on_screen_objects[i].blit_instead_of_draw = False
				self.on_screen_objects[i].alpha = 255
			self.on_screen_objects[i].draw(surf_temp)
		if self.rotation_degrees != 0.0:
			surf_temp = pygame.transform.rotate(surf_temp, -1.0 * self.rotation_degrees)

		# watchdog: make sure something is getting drawn to the display at least once every N frames
		if self.reset_on_no_drawing:
			if self.last_surface:
				ls_pixels = pygame.PixelArray(self.last_surface)
				current_pixels = pygame.PixelArray(surf_temp)
				done_checking = False
				found_different_pixel = False
				for x in range(0, len(ls_pixels)):
					if not done_checking:
						for y in range(0, len(ls_pixels[x])):
							if not done_checking:
								if ls_pixels[x][y] != current_pixels[x][y]:
									found_different_pixel = True
									done_checking = True
									break
				if found_different_pixel:
					self.unchanging_surface_count = 0
				else:
					self.unchanging_surface_count += 1				
					if self.unchanging_surface_count > self.max_unchanged_frames:
						print("Resetting current object manager due to lack of drawing")
						self.unchanging_surface_count = 0
						self.reset()
			
		#self.last_surface = surf_temp.copy()
		# use a scaled version of the surface to make comparison significantly faster on slow hardware
		self.last_surface = pygame.transform.smoothscale(surf_temp, (self.screen_width / self.last_surface_scaling_factor, self.screen_height / self.last_surface_scaling_factor))
		return surf_temp

	def get_spooky_colours(self):
		result = []
		# 80s vector white
		result.append((230, 255, 245))
		# Alien 3 computer green
		result.append((27, 255, 90))
		# ominous red
		result.append((200, 0, 0))
		# bright purple
		result.append((255, 0, 255))
		# bright blue
		result.append((0, 0, 255))
		# bright red
		result.append((255, 0, 0))
		# pale-ish blue
		result.append((54, 77, 255))
		# pale-ish blue
		result.append((91, 137, 255))
		return result
		
	def get_bright_colours(self):
		result = []
		# 80s vector white
		result.append((230, 255, 245))
		# Alien 3 computer green
		result.append((27, 255, 90))
		# bright purple
		result.append((255, 0, 255))
		# bright blue
		result.append((27, 90, 255))
		# bright red
		result.append((255, 0, 0))
		return result
		
	def get_highlight_colours(self):
		result = []
		#result.append((0, 255, 0))
		# Alien 3 computer green
		result.append((27, 255, 90))
		# ominous purple
		result.append((192, 0, 255))
		#result.append((255, 0, 255))
		# bright blue
		result.append((27, 90, 255))
		# ominous red
		result.append((200, 0, 0))
		#result.append((255, 0, 0))
		return result
		
	@staticmethod
	def get_colour_gradient(start_colour, end_colour, num_steps):
		result = []
		start_red = float(start_colour[0])
		start_green = float(start_colour[1])
		start_blue = float(start_colour[2])
		
		end_red = float(end_colour[0])
		end_green = float(end_colour[0])
		end_blue = float(end_colour[0])
		
		fns = float(num_steps)
		
		red_step = (start_red - end_red) / fns
		green_step = (start_green - end_green) / fns
		blue_step = (start_blue - end_blue) / fns
				
		for i in range(0, num_steps - 1):
			r = max(0, min(255, int(round(start_red + (red_step * i)))))
			g = max(0, min(255, int(round(start_green + (green_step * i)))))
			b = max(0, min(255, int(round(start_blue + (blue_step * i)))))
			#print("Debug: colour is %s, %s, %s" % (r, g, b))
			result.append(pygame.Color(r, g, b))
		
		# So that the last colour really is correct and not off by 1 due to rounding
		result.append(end_colour)
		
		#print("Gradient: %s" % (result))
		
		return result
				
class fast_hex_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(fast_hex_white, self).__init__(screen, screen_width, screen_height)
		self.num_boxes = 17
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
	
		self.box_min_size = 50
		self.box_max_size = 500
		
		self.frame_num = 0
		self.max_frame = 500
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		for i in range(0, self.num_boxes):
			new_obj = polygon(6)
			box_size = random.randint(self.box_min_size, self.box_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (self.screen_width / 2, self.screen_height / 2)
			new_obj.angular_velocity = random.randint(-10, 10)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = 1
			initial_y_velocity = 0
			if i % 2 == 0:
				initial_x_velocity = 0
			if i % 3 == 0:
				initial_y_velocity = 0
			if i % 4 == 0:
				initial_x_velocity = -1
				initial_y_velocity = 0
			if i % 5 == 0:
				initial_x_velocity = 0
				initial_y_velocity = -1
			if i % 6 == 0:
				initial_x_velocity = 0
				initial_y_velocity = 0
			#if i % 8 == 0:
			#	initial_x_velocity = -1
			#	initial_y_velocity = -1
			if i % 8 == 0:
				initial_x_velocity = -1
				initial_y_velocity = 0
			
			initial_x_velocity *= random.randint(1, 10)
			initial_y_velocity *= random.randint(1, 10)
				
			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)

class fast_polygon_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(fast_polygon_white, self).__init__(screen, screen_width, screen_height)
		self.num_polygons = 17
		self.min_num_sides = 3
		self.max_num_sides = 17
		self.num_sides = self.min_num_sides
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
		# +/- this amount
		self.max_initial_velocity = 5
	
		self.polygon_min_size = 50
		self.polygon_max_size = 500
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		self.num_sides = random.randint(self.min_num_sides, self.max_num_sides)
		for i in range(0, self.num_polygons):
			new_obj = polygon(self.num_sides)
			box_size = random.randint(self.polygon_min_size, self.polygon_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (self.screen_width / 2, self.screen_height / 2)
			new_obj.angular_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			initial_y_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)

			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
		

class slow_polygon_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygon_white, self).__init__(screen, screen_width, screen_height)
		self.num_polygons = 17
		self.min_num_sides = 3
		self.max_num_sides = 17
		self.num_sides = self.min_num_sides
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
		# +/- this amount
		self.max_initial_velocity = 5
		
	
		self.polygon_min_size = 5
		self.polygon_max_size = 50
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		self.num_sides = random.randint(self.min_num_sides, self.max_num_sides)
		for i in range(0, self.num_polygons):
			new_obj = polygon(self.num_sides)
			box_size = random.randint(self.polygon_min_size, self.polygon_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (random.randint(0, self.screen_width), random.randint(0, self.screen_height))
			new_obj.angular_velocity = 0
			new_obj.colour = self.colour_foreground
			initial_x_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			initial_y_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)

			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
			
class slow_polygon_rotating_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygon_rotating_white, self).__init__(screen, screen_width, screen_height)
		self.num_polygons = 17
		#self.num_polygons = 100
		self.min_num_sides = 3
		self.max_num_sides = 17
		self.num_sides = self.min_num_sides
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
		# +/- this amount
		self.max_initial_velocity = 5
		
	
		self.polygon_min_size = 5
		self.polygon_max_size = 50
		#self.polygon_min_size = 1
		#self.polygon_max_size = 10
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		self.num_sides = random.randint(self.min_num_sides, self.max_num_sides)
		for i in range(0, self.num_polygons):
			new_obj = polygon(self.num_sides)
			box_size = random.randint(self.polygon_min_size, self.polygon_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (random.randint(0, self.screen_width), random.randint(0, self.screen_height))
			new_obj.angular_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			new_obj.angular_velocity = 1
			new_obj.rotation_degrees = random.randint(0, 360)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			initial_y_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)

			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
			
class slow_polygrams_rotating_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygrams_rotating_white, self).__init__(screen, screen_width, screen_height)
		self.num_polygons = 3
		self.min_num_sides = 5
		self.max_num_sides = 19

		# +/- this amount
		self.max_initial_velocity = 1
			
		self.polygon_min_size = 25
		self.polygon_max_size = 500
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		self.num_sides = random.randint(self.min_num_sides, self.max_num_sides)
		for i in range(0, self.num_polygons):
			new_obj = polygram(self.num_sides)
			box_size = random.randint(self.polygon_min_size, self.polygon_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (random.randint(0, self.screen_width), random.randint(0, self.screen_height))
			new_obj.angular_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			#new_obj.angular_velocity = 1
			new_obj.rotation_degrees = random.randint(0, 360)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			initial_y_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)

			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
			
class slow_polygram_rotating_white(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygram_rotating_white, self).__init__(screen, screen_width, screen_height)
		self.num_polygons = 1
		self.min_num_sides = 5
		#self.max_num_sides = 19
		#self.max_num_sides = 47
		self.max_num_sides = 31
		self.mandala_mode_percentage = 25.0
		self.mandala_mode = False
		self.filled_polygon_percentage = 25.0
		self.filled_polygons = False
		
		# +/- this amount
		self.max_initial_velocity = 2
			
		self.polygon_min_size = 100
		self.polygon_max_size = 500
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
	def reset(self):
		mandala_percent = random.uniform(0.0, 100.0)
		self.mandala_mode = False
		if mandala_percent <= self.mandala_mode_percentage:
			self.mandala_mode = True
		fill_polygon_percent = random.uniform(0.0, 100.0)
		self.filled_polygons = False
		if fill_polygon_percent <= self.filled_polygon_percentage:
			self.filled_polygons = True
		self.current_settings_start_time = time.time()
		self.on_screen_objects = []
		self.num_sides = random.randint(self.min_num_sides, self.max_num_sides)
		for i in range(0, self.num_polygons):
			new_obj = polygram(self.num_sides)
			new_obj.mandala_mode = self.mandala_mode
			new_obj.fill_polygon = self.filled_polygons
			box_size = random.randint(self.polygon_min_size, self.polygon_max_size)
			new_obj.size = (box_size, box_size)
			new_obj.position = (random.randint(0, self.screen_width), random.randint(0, self.screen_height))
			new_obj.angular_velocity = 0
			#new_obj.angular_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			max_neg = self.max_initial_velocity * -1000
			max_pos = self.max_initial_velocity * 1000
			new_obj.angular_velocity = float(random.randint(max_neg, max_pos)) / 1000.0
			#while new_obj.angular_velocity == 0:
			#	new_obj.angular_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			#new_obj.angular_velocity = 1
			new_obj.rotation_degrees = random.randint(0, 360)
			new_obj.colour = self.colour_foreground
			initial_x_velocity = 0
			initial_y_velocity = 0
			initial_x_velocity = float(random.randint(max_neg, max_pos)) / 1000.0
			initial_y_velocity = float(random.randint(max_neg, max_pos)) / 1000.0
			# while initial_x_velocity == 0:
				# initial_x_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)
			# while initial_y_velocity == 0:
				# initial_y_velocity = random.randint(self.max_initial_velocity * -1, self.max_initial_velocity)

			new_obj.velocity = (initial_x_velocity, initial_y_velocity)
			
			self.on_screen_objects.append(new_obj)
			
class slow_polygram_rotating_coloured(slow_polygram_rotating_white):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygram_rotating_coloured, self).__init__(screen, screen_width, screen_height)
		self.colours = self.get_spooky_colours()
		
	def reset(self):
		super(slow_polygram_rotating_coloured, self).reset()
		colour_num = random.randint(0, (len(self.colours) - 1))
		new_colour = self.colours[colour_num]
		self.colour_foreground = new_colour
		for i in range(0, len(self.on_screen_objects)):
			self.on_screen_objects[i].colour_foreground = new_colour
			
class slow_polygrams_rotating_coloured(slow_polygram_rotating_coloured):
	def __init__(self, screen, screen_width, screen_height):
		super(slow_polygrams_rotating_coloured, self).__init__(screen, screen_width, screen_height)
		self.min_num_polygrams = 3
		#self.max_num_polygrams = 31
		self.max_num_polygrams = 17
		self.num_polygons = 7
		
	def reset(self):
		self.num_polygons = random.randint(self.min_num_polygrams, self.max_num_polygrams)
		super(slow_polygrams_rotating_coloured, self).reset()
		if self.filled_polygons:
			self.alpha = 96
		else:
			self.alpha = -1
			
class fast_polygrams_rotating_coloured(slow_polygrams_rotating_coloured):
	def __init__(self, screen, screen_width, screen_height):
		super(fast_polygrams_rotating_coloured, self).__init__(screen, screen_width, screen_height)
		self.max_initial_velocity = 30
		self.polygon_min_size = 25
		self.polygon_max_size = 500
		self.num_polygons = 7
		
	def reset(self):
		self.num_polygons = random.randint(self.min_num_polygrams, self.max_num_polygrams)
		super(slow_polygrams_rotating_coloured, self).reset()
		if self.filled_polygons:
			self.alpha = 96
		else:
			self.alpha = -1

class hexgrid_1(object_manager):

	def __init__(self, screen, screen_width, screen_height):
		super(hexgrid_1, self).__init__(screen, screen_width, screen_height)
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		self.colour_foreground_alt = (255, 0, 255)
		self.hexagon_size = 20
		self.hexagon_size_with_buffer = self.hexagon_size + 2
		#self.hex_grid_height	= (self.screen_width / (self.hexagon_size_with_buffer)) + 2
		#self.hex_grid_width = (self.screen_height / (self.hexagon_size_with_buffer)) + 2
		self.hex_grid_height	= 25
		self.hex_grid_width = 12
		#self.colour_flip_percent = 7
		self.colour_flip_percent = 1
		self.autoreset_interval_seconds = 0.0
		
	def reset(self):
		self.on_screen_objects = []
		for y in range(0, self.hex_grid_height):
			for x in range(0, self.hex_grid_width):
				new_obj = polygon(6)
				new_obj.size = (self.hexagon_size, self.hexagon_size)
#				if y % 2 == 1:
#					new_obj.position = (self.screen_width - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)) + self.hexagon_size_with_buffer)
#				else:
#					new_obj.position = (self.screen_width - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)))
				if y % 2 == 1:
					new_obj.position = ((self.screen_width + self.hexagon_size_with_buffer) - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)) + self.hexagon_size_with_buffer)
				else:
					new_obj.position = ((self.screen_width + self.hexagon_size_with_buffer) - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)))
				new_obj.angular_velocity = 0
				#new_obj.rotation_degrees = 90
				new_obj.velocity = (0, 0)
				new_obj.colour = colour_foreground
				self.on_screen_objects.append(new_obj)
				
	def main_loop(self):
		super(hexgrid_1, self).main_loop() # in Python 2 use super(D, self).main_loop()
		# randomly inject alternate-coloured objects and have them "infect" adjacent cells
		for i in range(0, len(self.on_screen_objects)):
			cell_list = []
			if self.on_screen_objects[i].colour == self.colour_foreground_alt:
				cell_list.append(i - self.hex_grid_width)
				cell_list.append(i - self.hex_grid_width + 1)
				cell_list.append(i - 1)
				cell_list.append(i + 1)
				cell_list.append(i + self.hex_grid_width)
				cell_list.append(i + self.hex_grid_width - 1)
			for c in cell_list:
				if c >= 0 and c < len(self.on_screen_objects):
					self.on_screen_objects[i].colour = self.colour_foreground_alt
			self.on_screen_objects[i].colour = self.colour_foreground
			

			if random.randint(0, 100) <= self.colour_flip_percent:
				self.on_screen_objects[i].colour = self.colour_foreground_alt

# this is super-interesting, don't change it
class hexgrid_2(object_manager):

	def __init__(self, screen, screen_width, screen_height):
		super(hexgrid_2, self).__init__(screen, screen_width, screen_height)
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		self.colour_foreground_alt = (255, 0, 255)
		self.hexagon_size = 20
		self.buffer_size = 2
		self.hexagon_size_with_buffer = self.hexagon_size + self.buffer_size
		#self.hexagon_size = 50
		#self.hexagon_size_with_buffer = self.hexagon_size + 20
		self.hex_grid_height = 25
		self.hex_grid_width = 12
		#self.hex_grid_height = 4
		#self.hex_grid_width = 2
		self.colour_flip_percent = 0.05
		#self.autoreset_interval_seconds = 0.0
		self.autoreset_interval_seconds = 31.337
		self.cell_values = []
		self.max_cell_value = 255
		# 16 = reasonably mid-paced
		self.decay_amount = 8
		self.min_decay_amount = 32
		self.max_decay_amount = 64
		#self.propagation_threshold = self.max_cell_value - (1 * self.decay_amount)
		self.propagation_threshold = 64
		self.adjacent_cell_boost = 32
		#self.adjacent_cell_boost = 8
		self.reset_cell_values()
		#self.colour_gradient = object_manager.get_colour_gradient(self.colour_foreground, self.colour_foreground_alt, 256)
		self.set_random_gradient()
		self.reset_on_no_drawing = False
		self.reset_on_no_objects = False
		self.cell_adjacency_mode = 1
		self.fill_cells = True
		#self.fill_cells = False
		self.fill_cells_mode_percentage = 66.6
		self.on_screen_objects = []
		
	def reset_cell_values(self):
		grid_size = self.hex_grid_width * self.hex_grid_height
		if len(self.cell_values) != grid_size:
			self.cell_values = []
			for i in range(0, grid_size):
				self.cell_values.append(0)
		else:
			for i in range(0, grid_size):
				self.cell_values[i] = 0

	def set_random_gradient(self):
		gradient_num = random.randint(0, 6)

		step1 = []
		step2 = []
		step3 = []
		step4 = []
		step5 = []
		
		# if gradient_num == 0:
			# # black to purple
			# self.colour_gradient = object_manager.get_colour_gradient((0, 0, 0), (255, 0, 255), 256)
		# #if gradient_num == 1:
			# # 80s vector white to purple
			# #self.colour_gradient = object_manager.get_colour_gradient((230, 255, 245), (255, 0, 255), 256)
		# #	self.colour_gradient = object_manager.get_colour_gradient((230, 255, 245), (0, 0, 0), 256)
		# if gradient_num == 2:
			# # blue to purple
			# #self.colour_gradient = object_manager.get_colour_gradient((0, 0, 255), (255, 0, 255), 256)
			# self.colour_gradient = object_manager.get_colour_gradient((0, 0, 0), (230, 255, 245), 256)
		# if gradient_num == 3:
			# # black to blue
			# self.colour_gradient = object_manager.get_colour_gradient((0, 0, 0), (0, 0, 255), 256)
		# if gradient_num == 4:
			# # black to green
			# self.colour_gradient = object_manager.get_colour_gradient((0, 0, 0), (0, 255, 0), 256)
		# if gradient_num == 5:
			# # black to red
			# self.colour_gradient = object_manager.get_colour_gradient((0, 0, 0), (255, 0, 0), 256)
		if gradient_num == 0:
			# green/blue/purple
			step1 = object_manager.get_colour_gradient((0, 0, 0), (0, 255, 0), 96)
			step2 = object_manager.get_colour_gradient((0, 255, 0), (0, 0, 255), 96)
			step3 = object_manager.get_colour_gradient((0, 0, 255), (255, 0, 255), 64)
		if gradient_num == 1:
			# flame
			#step1 = object_manager.get_colour_gradient((0, 0, 0), (255, 0, 0), 128)
			#step2 = object_manager.get_colour_gradient((255, 0, 0), (255, 255, 0), 64)
			#step3 = object_manager.get_colour_gradient((255, 255, 0), (255, 255, 255), 64)
			step1 = object_manager.get_colour_gradient((0, 0, 0), (70, 70, 70), 64)
			step2 = object_manager.get_colour_gradient((70, 70, 70), (255, 0, 0), 96)
			step3 = object_manager.get_colour_gradient((255, 0, 0), (255, 255, 0), 64)
			step4 = object_manager.get_colour_gradient((255, 255, 0), (255, 255, 255), 32)
		if gradient_num == 2:
			# green/cyan/vector white
			step1 = object_manager.get_colour_gradient((0, 0, 0), (0, 255, 0), 64)
			step2 = object_manager.get_colour_gradient((0, 255, 0), (0, 255, 255), 64)
			step3 = object_manager.get_colour_gradient((0, 255, 255), (230, 255, 245), 128)
		if gradient_num == 99:
			# birthday cake with sprinkles
			step1 = object_manager.get_colour_gradient((255, 0, 255), (0, 0, 255), 64)
			step2 = object_manager.get_colour_gradient((0, 0, 255), (255, 64, 64), 96)
			step3 = object_manager.get_colour_gradient((255, 190, 190), (255, 255, 0), 64)
			step4 = object_manager.get_colour_gradient((0, 255, 255), (255, 255, 255), 32)
		if gradient_num == 98:
			# new-style thermal imager
			step1 = object_manager.get_colour_gradient((0, 0, 0), (129, 29, 167), 64)
			step2 = object_manager.get_colour_gradient((129, 29, 167), (223, 109, 60), 64)
			step3 = object_manager.get_colour_gradient((223, 109, 60), (253, 205, 7), 64)
			step4 = object_manager.get_colour_gradient((253, 205, 7), (251, 228, 85), 32)
			step5 = object_manager.get_colour_gradient((251, 228, 85), (255, 255, 255), 32)
		if gradient_num == 3:
			# black/blue/white
			step1 = object_manager.get_colour_gradient((0, 0, 0), (0, 0, 255), 128)
			step2 = object_manager.get_colour_gradient((0, 0, 255), (230, 255, 245), 128)
		if gradient_num == 4:
			# black/green/white
			step1 = object_manager.get_colour_gradient((0, 0, 0), (0, 255, 0), 128)
			step2 = object_manager.get_colour_gradient((0, 255, 0), (230, 255, 245), 128)
		if gradient_num == 5:
			# black/purple/white
			step1 = object_manager.get_colour_gradient((0, 0, 0), (128, 0, 255), 128)
			step2 = object_manager.get_colour_gradient((128, 0, 255), (230, 255, 245), 128)
		if gradient_num == 6:
			# black/red/white
			step1 = object_manager.get_colour_gradient((0, 0, 0), (255, 0, 0), 128)
			step2 = object_manager.get_colour_gradient((255, 0, 0), (230, 255, 245), 128)
			
		for i in range(0, len(step2)):
			step1.append(step2[i])
		for i in range(0, len(step3)):
			step1.append(step3[i])
		for i in range(0, len(step4)):
			step1.append(step4[i])
		for i in range(0, len(step5)):
			step1.append(step5[i])
		self.colour_gradient = step1

	def set_random_parameters(self):
		param_set_num = random.randint(0, 5)
		#param_set_num = 5
		if param_set_num == 0:
			self.colour_flip_percent = 0.05
			self.decay_amount = 8
			self.propagation_threshold = 64
			self.adjacent_cell_boost = 32
		if param_set_num == 1:
			self.colour_flip_percent = 1.0
			self.decay_amount = 8
			self.propagation_threshold = 64
			self.adjacent_cell_boost = 32
		if param_set_num == 2:
			self.colour_flip_percent = 0.05
			self.decay_amount = 32
			self.propagation_threshold = 64
			self.adjacent_cell_boost = 64
		if param_set_num == 3:
			self.colour_flip_percent = 0.05
			self.decay_amount = 64
			self.propagation_threshold = 128
			self.adjacent_cell_boost = 64
		if param_set_num == 4:
			self.colour_flip_percent = 0.05
			self.decay_amount = 64
			self.propagation_threshold = 128
			self.adjacent_cell_boost = 192
		if param_set_num == 5:
			self.colour_flip_percent = 0.05
			self.decay_amount = 100
			self.propagation_threshold = 128
			self.adjacent_cell_boost = 192
			
	def reset(self):
		#self.decay_amount = random.randint(self.min_decay_amount, self.max_decay_amount)
		self.set_random_gradient()
		self.set_random_parameters()
		self.reset_cell_values()
		self.cell_adjacency_mode = random.randint(0, 3)
		fill_cell_val = random.uniform(0.0, 100.0)
		if fill_cell_val <= self.fill_cells_mode_percentage:
			self.fill_cells = True
		else:
			self.fill_cells = False
		#self.cell_adjacency_mode = 3
		#self.on_screen_objects = []
		create_new = False
		if len(self.on_screen_objects) == 0:
			create_new = True
		for y in range(0, self.hex_grid_height):
			for x in range(0, self.hex_grid_width):
				new_obj = None
				oso_number = (y * self.hex_grid_width) + x
				if create_new:
					new_obj = polygon(6)
					#new_obj = polygon(9)
					#new_obj = polygon(4)
				else:
					new_obj = self.on_screen_objects[oso_number]
				if self.fill_cells:
					new_obj.fill_polygon = True
				else:
					new_obj.fill_polygon = False
				new_obj.size = (self.hexagon_size, self.hexagon_size)
#				if y % 2 == 1:
#					new_obj.position = (self.screen_width - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)) + self.hexagon_size_with_buffer)
#				else:
#					new_obj.position = (self.screen_width - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)))
				if y % 2 == 1:
					#new_obj.position = ((self.screen_width + self.hexagon_size_with_buffer) - ((1 + y) * (self.hexagon_size_with_buffer * 1.5)), (self.hexagon_size_with_buffer * 2 * (x)) + self.hexagon_size_with_buffer)
					new_obj.position = ((self.screen_width + self.hexagon_size_with_buffer) - ((1 + y) * (self.hexagon_size_with_buffer * 1.7)), (self.hexagon_size_with_buffer * 2.0 * (x) + self.hexagon_size_with_buffer))
				else:
					new_obj.position = ((self.screen_width + self.hexagon_size_with_buffer) - ((1 + y) * (self.hexagon_size_with_buffer * 1.7)), (self.hexagon_size_with_buffer * 2.0 * (x)))
				new_obj.angular_velocity = 0.0
				#new_obj.rotation_degrees = 90
				new_obj.velocity = (0.0, 0.0)
				new_obj.colour = self.colour_foreground
				if create_new:
					self.on_screen_objects.append(new_obj)
				else:
					self.on_screen_objects[oso_number] = new_obj
	
	def get_wrapped_cell_index(self, cell_index):
		result = cell_index
		while result < 0:
			result += len(self.cell_values)
		while result >= len(self.cell_values):
			result -= len(self.cell_values)
		#print("Debug: %s wrapped is %s" % (cell_index, result))
		return result
	
	def main_loop(self):
		super(hexgrid_2, self).main_loop() # in Python 2 use super(D, self).main_loop()
		# randomly inject alternate-coloured objects and have them "infect" adjacent cells
		#for i in range(0, len(self.on_screen_objects)):
		#	colour_index = random.randint(0, 255)
		#	new_colour = self.colour_gradient[colour_index]
		#	self.on_screen_objects[i].colour = new_colour
		#return
		cell_value_changes = []
		for i in range(0, len(self.cell_values)):
			cell_value_changes.append(0)
		for i in range(0, len(self.on_screen_objects)):
			if self.cell_values[i] > self.max_cell_value:
				self.cell_values[i] = self.max_cell_value
			if self.cell_values[i] < 0:
				self.cell_values[i] = 0
			num_adjacent_cells = 0
			adjacent_cell_indices = []
			# east and west of the current cell
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i-1))
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i+1))
			# northeast and northwest of the current cell
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i-(self.hex_grid_width - 1)))
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i-(self.hex_grid_width)))
			# southeast and southwest of the current cell
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i+(self.hex_grid_width - 1)))
			adjacent_cell_indices.append(self.get_wrapped_cell_index(i+(self.hex_grid_width)))
			for aci in adjacent_cell_indices:
				if self.cell_values[aci] >= self.propagation_threshold:
					num_adjacent_cells += 1
			#if num_adjacent_cells > 0:
			#	print("Debug: num adjacent cells is %s" % (num_adjacent_cells))
			if self.cell_adjacency_mode == 0:
				if num_adjacent_cells > 0 and num_adjacent_cells < 3:
					cell_value_changes[i] += self.adjacent_cell_boost

				if num_adjacent_cells > 2 and num_adjacent_cells < 5:
					cell_value_changes[i] += (self.adjacent_cell_boost / 2)
					#cell_value_changes[i] += self.adjacent_cell_boost
			if self.cell_adjacency_mode == 1:
				if num_adjacent_cells > 0 and num_adjacent_cells < 3:
					cell_value_changes[i] += self.adjacent_cell_boost
				if num_adjacent_cells > 2 and num_adjacent_cells < 7:
					cell_value_changes[i] -= self.adjacent_cell_boost
			if self.cell_adjacency_mode == 2:
				if num_adjacent_cells > 0 and num_adjacent_cells < 4:
					cell_value_changes[i] += self.adjacent_cell_boost
				if num_adjacent_cells > 3 and num_adjacent_cells < 7:
					cell_value_changes[i] -= self.adjacent_cell_boost
			if self.cell_adjacency_mode == 3:
				if num_adjacent_cells > 0 and num_adjacent_cells < 2:
					cell_value_changes[i] += self.adjacent_cell_boost
				if num_adjacent_cells > 1 and num_adjacent_cells < 4:
					cell_value_changes[i] += (self.adjacent_cell_boost / 2)
				if num_adjacent_cells > 3 and num_adjacent_cells < 5:
					cell_value_changes[i] -= self.adjacent_cell_boost
				if num_adjacent_cells > 4 and num_adjacent_cells < 6:
					cell_value_changes[i] -= (self.adjacent_cell_boost * 2)
				if num_adjacent_cells > 5:
					cell_value_changes[i] -= (self.adjacent_cell_boost * 100)
		
			#if num_adjacent_cells > 4 and num_adjacent_cells < 6:
			#	cell_value_changes[i] -= (self.adjacent_cell_boost / 2)

			#if num_adjacent_cells > 5:
			#	cell_value_changes[i] -= (self.adjacent_cell_boost)
			
			# cell_list = []
			# if self.cell_values[i] >= self.propagation_threshold:
				# cell_list.append(i - self.hex_grid_width)
				# cell_list.append(i - self.hex_grid_width + 1)
				# cell_list.append(i - 1)
				# cell_list.append(i + 1)
				# cell_list.append(i + self.hex_grid_width)
				# cell_list.append(i + self.hex_grid_width - 1)
			# for c in cell_list:
				# if c >= 0 and c < len(self.on_screen_objects):
					# #self.cell_value_changes[c] += self.adjacent_cell_boost
					# self.cell_value_changes[c] += self.adjacent_cell_boost
					# if self.cell_value_changes[c] > self.max_cell_value:
						# self.cell_value_changes[c] = self.max_cell_value
					# if self.cell_value_changes[c] < 0:
						# self.cell_value_changes[c] = 0
			#print("Debug: current cell value is %s" % (self.cell_values[i]))
			#print("Debug: len(self.cell_values) == %s" % (len(self.cell_values)))
			#print("Debug: len(self.on_screen_objects) == %s" % (len(self.on_screen_objects)))
			#print("Debug: len(self.colour_gradient) == %s" % (len(self.colour_gradient)))
			colour_index = self.cell_values[i]
			#colour_as_string = "index %s: %s" % (colour_index, self.colour_gradient[colour_index])
			#print("Debug: setting cell colour to %s" % (colour_as_string))
			new_colour = (0, 0, 0)
			try:
				new_colour = self.colour_gradient[colour_index]
			except IndexError as e:
				print("Unable to get colour for index %s, length of gradient array is %s" % (colour_index, len(self.colour_gradient)))
			self.on_screen_objects[i].colour = (new_colour)
			if (float(random.randint(0, 10000)) / 100.0) <= self.colour_flip_percent:
				self.cell_values[i] = self.max_cell_value + self.decay_amount
			self.cell_values[i] -= self.decay_amount
		for i in range(0, len(cell_value_changes)):
			self.cell_values[i] += cell_value_changes[i]
			if self.cell_values[i] > self.max_cell_value:
				self.cell_values[i] = self.max_cell_value
			if self.cell_values[i] < 0:
				self.cell_values[i] = 0

class sweeping_lines(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(sweeping_lines, self).__init__(screen, screen_width, screen_height)
		self.min_num_lines = 0
		self.max_num_lines = 23
		self.min_spacing = 25
		self.max_spacing = 500

		self.min_initial_velocity = 1
		self.max_initial_velocity = 1
		# +/-
		self.max_initial_angular_velocity = 3.0
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		
		self.origin_corner_UPPER_LEFT = 0
		self.origin_corner_UPPER_RIGHT = 1
		self.origin_corner_LOWER_RIGHT = 2
		self.origin_corner_LOWER_LEFT = 3
		
		self.edge_buffer_size = 10
		
		self.boundary_edge_buffer_size = self.edge_buffer_size * 100
		
		self.origin_corner = self.origin_corner_UPPER_LEFT
		
		self.colours = self.get_bright_colours()
		
		self.bounce_at_edge = True
		
		self.enable_rotation = False
		
	def reset(self):		
		self.on_screen_objects = []
		self.origin_corner = random.randint(0, 3)
		num_lines_axis_1 = random.randint(self.min_num_lines, self.max_num_lines)
		num_lines_axis_2 = random.randint(self.min_num_lines, self.max_num_lines)
		colour_num = random.randint(0, (len(self.colours) - 1))
		new_colour = self.colours[colour_num]
		self.colour_foreground = new_colour
		line_length = screen_width
		if screen_height > screen_width:
			line_length = screen_height
		if self.enable_rotation:
			line_length *= 5		
		else:
			line_length *= 1.25
		
		line_velocity = random.randint(self.min_initial_velocity, self.max_initial_velocity)
		line_separation = random.randint(self.min_spacing * -1, self.max_spacing)
		
		line_angular_velocity = 0.0
		if self.enable_rotation:
			line_angular_velocity = float(random.randint(-1 * self.max_initial_angular_velocity * 100, self.max_initial_angular_velocity * 100)) / 100
		
		initial_axis_1_velocity = line_velocity
		initial_axis_2_velocity = float(line_velocity) / (float(self.screen_height) / float(self.screen_width))
		
		base_axis_1_start_position = (-1 * self.edge_buffer_size, -1 * self.edge_buffer_size)
		base_axis_2_start_position = (-1 * self.edge_buffer_size, -1 * self.edge_buffer_size)
		
		line_spacing_value_axis_1 = -1 * line_separation
		line_spacing_value_axis_2 = -1 * line_separation
		
		# all of these are rotated 90 degrees
		# top to bottom, right to left
		if self.origin_corner == self.origin_corner_UPPER_LEFT:
			initial_axis_2_velocity *= -1
			line_spacing_value_axis_2 *= -1
			base_axis_2_start_position = (self.screen_width + self.edge_buffer_size, -1 * self.edge_buffer_size)
		
		# bottom to top, right to left
		if self.origin_corner == self.origin_corner_UPPER_RIGHT:
			initial_axis_1_velocity *= -1
			initial_axis_2_velocity *= -1
			line_spacing_value_axis_1 *= -1
			line_spacing_value_axis_2 *= -1
			base_axis_1_start_position = (-1 * self.edge_buffer_size, self.screen_height + self.edge_buffer_size)
			base_axis_2_start_position = (self.screen_width + self.edge_buffer_size, -1 * self.edge_buffer_size)
		
		# bottom to top, left to right
		if self.origin_corner == self.origin_corner_LOWER_RIGHT:
			initial_axis_1_velocity *= -1
			line_spacing_value_axis_1 *= -1
			base_axis_1_start_position = (-1 * self.edge_buffer_size, self.screen_height + self.edge_buffer_size)

		# top to bottom, left to right
		#if self.origin_corner == self.origin_corner_LOWER_LEFT:
		
			
		for i in range(0, num_lines_axis_1):
			# vertical (on 90-degree-rotated screen)
			new_obj_axis_1 = arbitrary_geometry()
			new_obj_axis_1.colour_foreground = new_colour
			new_obj_axis_1.points.append((0, 0))
			new_obj_axis_1.points.append((line_length, 0))
			new_obj_axis_1.angular_velocity = line_angular_velocity
			new_obj_axis_1.bounce_at_edge = self.bounce_at_edge
			new_obj_axis_1.velocity = (0, initial_axis_1_velocity)
			new_obj_axis_1.position = (base_axis_1_start_position[0], base_axis_1_start_position[1] + (i * line_spacing_value_axis_1))
			#new_obj_axis_1.position = (self.screen_width / 2, self.screen_height / 2)
			new_obj_axis_1.colour = self.colour_foreground
			self.on_screen_objects.append(new_obj_axis_1)
			
		for i in range(0, num_lines_axis_2):
			# horizontal (on 90-degree-rotated screen)			
			new_obj_axis_2 = arbitrary_geometry()
			new_obj_axis_2.colour_foreground = new_colour
			new_obj_axis_2.points.append((0, 0))
			new_obj_axis_2.points.append((0, line_length))
			new_obj_axis_2.angular_velocity = line_angular_velocity
			new_obj_axis_2.bounce_at_edge = self.bounce_at_edge
			new_obj_axis_2.velocity = (initial_axis_2_velocity, 0)
			new_obj_axis_2.position = (base_axis_2_start_position[0] + (i * line_spacing_value_axis_2), base_axis_2_start_position[1])
			#new_obj_axis_2.position = (self.screen_width / 2, self.screen_height / 2)
			new_obj_axis_2.colour = self.colour_foreground
			self.on_screen_objects.append(new_obj_axis_2)

			
	def main_loop(self):
		super(sweeping_lines, self).main_loop()
		still_alive = []
		
		for i in range(0, len(self.on_screen_objects)):
			#print("object %s: %s, %s" % (i, self.on_screen_objects[i].position[0], self.on_screen_objects[i].position[1]))
			is_still_alive = False
			if self.on_screen_objects[i].position[0] >= (-1 * self.boundary_edge_buffer_size):
				if self.on_screen_objects[i].position[0] <= (self.screen_width + self.boundary_edge_buffer_size):
					if self.on_screen_objects[i].position[1] >= (-1 * self.boundary_edge_buffer_size):
						if self.on_screen_objects[i].position[1] <= (self.screen_height + self.boundary_edge_buffer_size):
							still_alive.append(self.on_screen_objects[i])
							is_still_alive = True
			#if not is_still_alive:
				#print("\tobject is out of bounds")
							
		self.on_screen_objects = still_alive
		
		if len(self.on_screen_objects) < 1:
			self.reset()
			
class terminal_text_output(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(terminal_text_output, self).__init__(screen, screen_width, screen_height)
		self.rotation_degrees = 90.0
		self.terminal_width = 70
		self.terminal_height = 60
		self.terminal_width = 34
		self.terminal_height = 30
		self.text_lines = []
		self.line_buffer = []
		self.new_chars_only = []
		self.current_line_to_write_to = 0
		#self.min_characters_per_loop = 0.5
		self.min_characters_per_loop = 1.0
		#self.max_characters_per_loop = 4.0
		self.max_characters_per_loop = 3.0
		self.characters_per_loop = 0.5
		self.characters_per_loop_counter = 0.0
		self.font = None
		self.line_spacing = 2
		self.line_spacing_effective = 10
		self.loop_instead_of_scroll = True
		self.loop_iteration_count = 0
		self.loops_before_colour_reset = 1
		self.clear_ahead_lines = 5
		self.colour_new_chars_only = (0, 255, 0)
		#self.colour_new_chars_only = (255, 0, 0)
		#self.colour_new_chars_only = (255, 255, 255)
		self.blur_factor = 8
		self.blurred_alpha = 192
		self.autoreset_interval_seconds = 0.0
		self.display_seconds = 31.337 * 2.0
		self.reset_on_no_drawing = False
		self.reset_on_no_objects = False
		
		self.highlight_colours = self.get_highlight_colours()
		
		self.on_screen_objects = [ arbitrary_geometry() ]
		self.current_line_buffer = None
		
	def add_text_to_buffer(self, new_text):
		for line in new_text.splitlines():
			new_line = line.rstrip()
			self.line_buffer.append(new_line)
			#print("Debug: added '%s' to buffer, which now contains %s lines" % (new_line, len(self.line_buffer)))
			
	def wrap_text_and_add_to_buffer(self, new_text):
		for line in new_text.splitlines():
			new_line = line.rstrip()
			while len(new_line) > self.terminal_width:
				temp = new_line[0:self.terminal_width]
				break_pos = self.terminal_width
				if " " in temp:
					break_pos = temp.rfind(" ")
				self.line_buffer.append(new_line[0:break_pos])
				new_line = new_line[break_pos:].strip()
			self.line_buffer.append(new_line)
		
	def set_font(self, font_file, size):
		self.font = pygame.font.Font(font_file, size)
		self.line_spacing_effective = size + self.line_spacing
		
	def reset(self):		
		self.text_lines = []
		for i in range(0, self.terminal_height):
			self.text_lines.append("")
		self.line_buffer = []
		self.current_line_buffer = None
		self.pick_new_options()
		
	def pick_new_options(self):
		colour_num = random.randint(0, (len(self.highlight_colours) - 1))
		new_colour = self.highlight_colours[colour_num]
		self.colour_new_chars_only = new_colour
		self.characters_per_loop = float(random.randint(math.floor(self.min_characters_per_loop * 10), math.floor(self.max_characters_per_loop * 10))) / 10
			
	def main_loop(self):
		#super(terminal_text_output, self).main_loop()
		
		#print("Debug: buffer length: %s lines" % (len(self.line_buffer)))
		
		# scroll existing text, if necessary
		
		#if self.loop_instead_of_scroll:
		#	if self.current_line_to_write_to > self.terminal_height - 1:
		#		self.current_line_to_write_to = 0
		#		self.text_lines[self.current_line_to_write_to] = ""
		#else:
		if not self.loop_instead_of_scroll:
			if self.current_line_to_write_to >= self.terminal_height - 1:
				#print("Debug: scrolling display by one line")
				if not self.current_line_buffer:
					if len(self.line_buffer) > 0:
						self.current_line_to_write_to = len(self.text_lines) - 1
						for i in range(0, self.current_line_to_write_to):
							self.text_lines[i] = self.text_lines[i + 1]
						self.text_lines[self.current_line_to_write_to] = ""
					
		# grab the next line if the current buffer is empty
		if not self.current_line_buffer:
			if len(self.line_buffer) > 0:
				#print("Debug: getting next line")
				self.current_line_buffer = self.line_buffer[0]
				temp_buffer = []
				if len(self.line_buffer) > 1:
					for i in range(1, len(self.line_buffer)):
						temp_buffer.append(self.line_buffer[i])
				self.line_buffer = temp_buffer
				self.current_line_to_write_to += 1
				if self.loop_instead_of_scroll:
					if self.clear_ahead_lines > 0:
						for i in range(0, self.clear_ahead_lines):
							wrapped_line_number = (self.current_line_to_write_to + 1 + i) % self.terminal_height
							self.text_lines[wrapped_line_number] = ""
					if self.current_line_to_write_to > self.terminal_height - 1:
						self.current_line_to_write_to = 0
						self.text_lines[self.current_line_to_write_to] = ""
						self.loop_iteration_count += 1
						if self.loop_iteration_count > self.loops_before_colour_reset:
							self.loop_iteration_count = 0
							self.pick_new_options()
				else:
					if self.current_line_to_write_to >= self.terminal_height - 1:
						self.current_line_to_write_to = self.terminal_height - 1
				self.text_lines[self.current_line_to_write_to] = ""
			#else:
				#print("Debug: no lines in buffer")
				
		if self.current_line_buffer:
			self.characters_per_loop_counter += self.characters_per_loop
			#print("Debug: self.characters_per_loop_counter: %s" % (self.characters_per_loop_counter))
			characters_to_add = 0
			new_chars = ""
			if self.characters_per_loop_counter >= 1.0:
				num_chars = int(math.floor(self.characters_per_loop_counter))
				if num_chars > len(self.current_line_buffer):
					num_chars = len(self.current_line_buffer)
				self.characters_per_loop_counter -= num_chars
				if num_chars == len(self.current_line_buffer):
					new_chars = self.current_line_buffer
					self.current_line_buffer = None
				else:
					new_chars = self.current_line_buffer[0:num_chars]
					self.current_line_buffer = self.current_line_buffer[num_chars:]
				#print("Adding '%s' to line %s" % (new_chars, self.current_line_to_write_to))
				
				# update the "only new text" buffer (for visual cursor-like effects)
				self.new_chars_only = []
				if self.current_line_to_write_to > 0:
					for i in range(0, self.current_line_to_write_to):
						self.new_chars_only.append("")
				current_line_only_new = ""
				if len(self.text_lines[self.current_line_to_write_to]) > 0:
					for i in range(0, len(self.text_lines[self.current_line_to_write_to])):
						current_line_only_new += " "
				current_line_only_new += new_chars
				self.new_chars_only.append(current_line_only_new)
				for i in range(self.current_line_to_write_to + 1, self.terminal_height):
					self.new_chars_only.append("")
				self.text_lines[self.current_line_to_write_to] += new_chars

	def draw(self, surf):
		draw_text = True
		if not self.font:
			draw_text = False
		lines_with_text = 0
		for i in range(0, len(self.text_lines)):
			if self.text_lines[i].strip() != "":
				lines_with_text += 1
		if lines_with_text == 0:
			draw_text = False
		if draw_text:
			surf_temp = None
			if self.rotation_degrees != 0.0:
				surf_temp = pygame.transform.rotate(surf, self.rotation_degrees)
			else:
				surf_temp = surf.copy()
			for i in range(0, len(self.text_lines)):
				text_surface = self.font.render(self.text_lines[i], True, self.colour_foreground)
				text_rect = text_surface.get_rect()
				text_rect.topleft = (0, (self.line_spacing_effective * i))
				surf_temp.blit(text_surface, text_rect)
			# add new-character-only effects
			# get just the new text
			if self.new_chars_only[self.current_line_to_write_to].strip() != "":
				nco_text_surface = self.font.render(self.new_chars_only[self.current_line_to_write_to], True, self.colour_new_chars_only)
				nco_text_surface_rect = nco_text_surface.get_rect()
				nco_text_surface_rect.topleft = (0, (self.line_spacing_effective * self.current_line_to_write_to))
				blurred_width = nco_text_surface_rect.width / self.blur_factor
				if blurred_width < 1:
					blurred_width = 1
				blurred_height = nco_text_surface_rect.height / self.blur_factor
				if blurred_height < 1:
					blurred_height = 1
				nco_text_surface_blurred = pygame.transform.smoothscale(pygame.transform.smoothscale(nco_text_surface, (blurred_width, blurred_height)), (nco_text_surface_rect.width, nco_text_surface_rect.height))
				nco_text_surface_blurred.set_colorkey(colour_background)
				nco_text_surface_blurred.set_alpha(self.blurred_alpha)				
				# add the new text with effects to the existing text
				surf_temp.blit(nco_text_surface, nco_text_surface_rect)
				surf_temp.blit(nco_text_surface_blurred, nco_text_surface_rect)
			if self.rotation_degrees != 0.0:
				surf_temp = pygame.transform.rotate(surf_temp, -1.0 * self.rotation_degrees)
			return surf_temp
		return surf
		
class terminal_text_infinite_loop(terminal_text_output):
	def __init__(self, screen, screen_width, screen_height):
		super(terminal_text_infinite_loop, self).__init__(screen, screen_width, screen_height)
		self.base_text = ""
		super(terminal_text_infinite_loop, self).reset()
		
	def main_loop(self):
		super(terminal_text_infinite_loop, self).main_loop()
		if len(self.line_buffer) == 0:
			self.wrap_text_and_add_to_buffer(self.base_text)

	def reset(self):		
		self.pick_new_options()

class matrix_cursor(object):
	def __init__(self):
		self.position = -1.0
		self.speed = 1.0
		self.character = "X"
	
	def pick_new_char(self, character_list):
		rand_char_num = random.randint(0, len(character_list) - 1)
		self.character = character_list[rand_char_num:rand_char_num+1]
		
class matrix_code(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(matrix_code, self).__init__(screen, screen_width, screen_height)
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\￥ァアィイゥウェエォオカガキギクグケゲコゴサザシジスズセゼソゾタダチヂッツヅテデトドナニヌネノハバパヒビピフブプヘベペホボポマミムメモャヤュユョヨラリルレロヮワヰヱヲンヴヵヶヷヸヹヺ・ーヽヾヿАБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX'
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX'
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX־׀׃׆אבגדהוזחטיךכלםמןנסעףפץצקרשתװױײ׳״؆؇؈؉؊؋،؍؎؏؛؞؟ؠءآأؤإئابةتثجحخدذرزسشصضطظعغػؼؽؾؿـفقكلمنهوىي٠١٢٣٤٥٦٧٨٩٪٫٬٭ٮٯٱٲٳٵٶٷٸٹٺٻټٽپٿڀځڂڃڄڅچڇڈډڊڋڌڍڎڏڐڑڒړڔڕږڗژڙښڛڜڝڞڟڠڡڢڣڤڥڦڧڨکڪګڬڭڮگڰڱڲڳڴڵڶڷڸڹںڻڼڽھڿۀہۂۃۄۅۆۇۈۉۊۋیۍێۏېۑےۓ۔ە۞۩ۮۯ۰۱۲۳۴۵۶۷۸۹ۺۻۼ۽۾ۿ'
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX׆אבגדהוזחטכלם'
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\=≠≤≥АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX¥ǂƵƾƼŦƛƀΣΛΔΘΞΨΩΠｦｧｨｩｪｫｬｭｮｯｰｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝ'	
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\=≠АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX¥ǂƵƾƼŦƛƀΣΛΔΘΞΨΩΠｦｧｨｩｪｫｬｭｮｯｰｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝ'
		#self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\=≠АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX¥ǂƵŦƛƀΣΛΔΘΞΨΩΠｦｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝ'
		self.characters = u'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%&*?/\\=≠АБВГДЕЖЗИЙКЛМНОПРСТУХЦЧШЩЫЬЮЯX¥ǂƵŦƛƀΣΛΔΘΞΨΩΠｦｱｲｳｴｵｶｷｸｹｺｻｼｽｾｿﾀﾁﾂﾃﾄﾅﾆﾇﾈﾉﾊﾋﾌﾍﾎﾏﾐﾑﾒﾓﾔﾕﾖﾗﾘﾙﾚﾛﾜﾝ'
		self.rotation_degrees = 90.0
		self.terminal_width = 70
		self.terminal_height = 60
		self.terminal_width = 40
		self.terminal_height = 30
		self.text_lines = []
		self.new_chars_only = []
		for i in range(0, self.terminal_height):
			new_row = []
			new_row2 = []
			for x in range(0, self.terminal_width):
				new_row.append(" ")
				new_row2.append(" ")
			self.text_lines.append(new_row)
			self.new_chars_only.append(new_row2)
		self.font = None
		self.line_spacing = 2
		self.line_spacing_effective = 10
		self.loop_iteration_count = 0
		self.loops_before_colour_reset = 1
		self.clear_ahead_lines = 7
		self.colour_foreground = (230, 0, 13)
		self.colour_new_chars_only = (255, 255, 255)
		self.blur_factor = 8
		self.blurred_alpha = 192
		self.autoreset_interval_seconds = 31.337
		#self.min_cursor_speed = 0.1
		#self.min_cursor_speed = 0.01
		self.min_cursor_speed = 0.05
		#self.max_cursor_speed = 4.0
		#self.max_cursor_speed = 3.0
		self.max_cursor_speed = 2.0
		self.existing_character_change_percent = 10.0
		self.pride_mode_colours = []
		self.pride_mode_colours.append((255, 0, 0))
		self.pride_mode_colours.append((255, 128, 0))
		self.pride_mode_colours.append((255, 255, 0))
		self.pride_mode_colours.append((0, 255, 0))
		self.pride_mode_colours.append((0, 255, 255))
		self.pride_mode_colours.append((0, 0, 255))
		self.pride_mode_colours.append((255, 0, 255))
		self.pride_mode_counter = 0
		self.pride_mode_enabled = False
		self.reset_on_no_drawing = False
		self.reset_on_no_objects = False
		
		#self.max_cursor_speed = 0.1
		
		self.highlight_colours = self.get_highlight_colours()
		self.on_screen_objects = [ arbitrary_geometry() ]	
		self.cursors = []
		for i in range(0, self.terminal_width):
			new_cursor = matrix_cursor()
			self.cursors.append(new_cursor)
		
	def set_font(self, font_file, size):
		self.font = pygame.font.Font(font_file, size)
		self.line_spacing_effective = size + self.line_spacing
		
	def reset(self):
		self.current_settings_start_time = time.time()
		self.existing_character_change_percent = random.uniform(-2.0, 10.0)
		for i in range(0, len(self.cursors)):
			#self.cursors[i].position = -1.0
			self.cursors[i].position = 0.0
			self.cursors[i].speed = random.uniform(self.min_cursor_speed, self.max_cursor_speed)
			self.cursors[i].pick_new_char(self.characters)
		#self.text_lines = []
		for y in range(0, self.terminal_height):
			#new_row = []
			#new_row2 = []
			for x in range(0, self.terminal_width):
				self.text_lines[y][x] = " "
				self.new_chars_only[y][x] = " "
				#new_row.append(" ")
				#new_row2.append(" ")
			#self.text_lines.append(new_row)
			#self.new_chars_only.append(new_row2)
		self.pick_new_options()
		
	def pick_new_options(self):
		#colour_num = random.randint(0, len(self.highlight_colours))
		colour_num = random.randint(0, len(self.highlight_colours) - 1)
		if colour_num < len(self.highlight_colours):
			self.pride_mode_enabled = False
			new_colour = self.highlight_colours[colour_num]
			self.colour_foreground = new_colour
		else:
			self.pride_mode_enabled = True
			self.pride_mode_counter = 0
		#self.pride_mode_enabled = True

	def main_loop(self):
		if self.autoreset_interval_seconds > 0.0:
			elapsed = time.time() - self.current_settings_start_time
			if elapsed >= self.autoreset_interval_seconds:
				self.current_settings_start_time = time.time()
				self.reset()
		for i in range(0, len(self.cursors)):
			cursor_previous_row = int(math.floor(self.cursors[i].position))
			self.cursors[i].position += self.cursors[i].speed
			if self.cursors[i].position > self.terminal_height:
				#self.cursors[i].position = -1.0
				self.cursors[i].position = 0.0
				self.cursors[i].speed = random.uniform(self.min_cursor_speed, self.max_cursor_speed)
				#for y in range(0, len(self.text_lines)):
				#	self.text_lines[y][i] = " "
			if self.cursors[i].position >= 0:
				cursor_current_row = int(math.floor(self.cursors[i].position))
				if cursor_previous_row != cursor_current_row:
					self.text_lines[cursor_previous_row][i] = self.cursors[i].character
					if self.cursors[i].speed > 1.0:
						#backfill_steps = int(math.floor(self.cursors[i].speed)) - 1
						backfill_steps = cursor_current_row - cursor_previous_row
						if cursor_previous_row > cursor_current_row:
							backfill_steps = (cursor_current_row + self.terminal_height) - cursor_previous_row
						if backfill_steps > 0:
							for b in range(0, backfill_steps):
								backfill_pos = cursor_current_row - (1 + b)
								if backfill_pos < 0:
									backfill_pos += self.terminal_height
								#if backfill_pos == cursor_previous_row and self.text_lines[cursor_previous_row][i] != " ":
								#	break
								char_num = random.randint(0, (len(self.characters) - 1))
								self.text_lines[backfill_pos][i] = self.characters[char_num]
				self.cursors[i].pick_new_char(self.characters)
				for c in range(1, self.clear_ahead_lines + 1 + (int(math.floor(self.cursors[i].speed)) * 2)):
				#for c in range(1, self.clear_ahead_lines + 1):
					wrapped_row = (cursor_current_row + c) % self.terminal_height
					self.text_lines[wrapped_row][i] = " "
				
				for y in range(0, len(self.new_chars_only)):
					self.new_chars_only[y][i] = " "
				self.new_chars_only[cursor_current_row][i] = self.cursors[i].character
		# randomize existing characters
		if self.existing_character_change_percent > 0.0:
			for y in range(0, self.terminal_height):
				for x in range(0, self.terminal_width):
					if self.text_lines[y][x] != " ":
						r = random.uniform(0.0, 100.0)
						if r <= self.existing_character_change_percent:
							rand_char_num = random.randint(0, len(self.characters) - 1)
							self.text_lines[y][x] = self.characters[rand_char_num:rand_char_num+1]

	def draw(self, surf):
		draw_text = True
		if not self.font:
			draw_text = False

		if draw_text:
			surf_temp = None
			if self.rotation_degrees != 0.0:
				surf_temp = pygame.transform.rotate(surf, self.rotation_degrees)
			else:
				surf_temp = surf.copy()
			self.pride_mode_counter = 0
			for y in range(0, len(self.text_lines)):
				current_line = ""
				for x in range(0, len(self.text_lines[y])):
					current_line += self.text_lines[y][x]
				line_colour = self.colour_foreground
				if self.pride_mode_enabled:
					line_colour = self.pride_mode_colours[self.pride_mode_counter]
					self.pride_mode_counter += 1
					if self.pride_mode_counter >= len(self.pride_mode_colours):
						self.pride_mode_counter = 0
				text_surface = self.font.render(current_line, True, line_colour)
				
				text_rect = text_surface.get_rect()
				text_rect.topleft = (0, (self.line_spacing_effective * y))
				surf_temp.blit(text_surface, text_rect)
			for y in range(0, len(self.new_chars_only)):
				current_line = ""
				for x in range(0, len(self.new_chars_only[y])):
					current_line += self.new_chars_only[y][x]
				if current_line.strip() != "":					
					nco_text_surface = self.font.render(current_line, True, self.colour_new_chars_only)
					nco_text_surface_rect = nco_text_surface.get_rect()
					nco_text_surface_rect.topleft = (0, (self.line_spacing_effective * y))
					blurred_width = nco_text_surface_rect.width / self.blur_factor
					if blurred_width < 1:
						blurred_width = 1
					blurred_height = nco_text_surface_rect.height / self.blur_factor
					if blurred_height < 1:
						blurred_height = 1
					nco_text_surface_blurred = pygame.transform.smoothscale(pygame.transform.smoothscale(nco_text_surface, (blurred_width, blurred_height)), (nco_text_surface_rect.width, nco_text_surface_rect.height))
					nco_text_surface_blurred.set_colorkey(colour_background)
					nco_text_surface_blurred.set_alpha(self.blurred_alpha)				
					# add the new text with effects to the existing text
					surf_temp.blit(nco_text_surface, nco_text_surface_rect)
					surf_temp.blit(nco_text_surface_blurred, nco_text_surface_rect)
			if self.rotation_degrees != 0.0:
				surf_temp = pygame.transform.rotate(surf_temp, -1.0 * self.rotation_degrees)
			return surf_temp
		return surf

class sweeping_lines_no_bounce(sweeping_lines):
	def __init__(self, screen, screen_width, screen_height):
		super(sweeping_lines_no_bounce, self).__init__(screen, screen_width, screen_height)
		self.bounce_at_edge = False
		
class image_slideshow(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(image_slideshow, self).__init__(screen, screen_width, screen_height)
		self.rotation_degrees = 90.0
		self.image_paths = []
		self.current_queue = []
		self.current_image = None
		self.shuffle_queue = True
		self.seconds_per_image = 10.0
		#self.seconds_per_image = 5
		self.image_start_time = time.time()
		self.on_screen_objects = []
		self.current_line_buffer = None
		self.reset_on_no_drawing = False
		self.reset_on_no_objects = False
		self.enable_chromakey = False
		self.add_post_blur = False
		self.autoreset_interval_seconds = 0.0
		
	def reset(self):		
		self.current_queue = []
		#self.elapsed_millisecond_counter = 0
		map_index = []
		if self.shuffle_queue:
			while len(map_index) < len(self.image_paths):
				next_index = random.randint(0, len(self.image_paths) - 1)
				if next_index not in map_index:
					map_index.append(next_index)
			#print("Debug: map_index '%s'" % (map_index))
		else:
			for i in range(0, len(self.image_paths)):
				map_index.append(i)
		for i in range(0, len(self.image_paths)):
			new_img = on_screen_bitmap()
			bitmap_path = self.image_paths[map_index[i]]
			#print("Debug: set image path to = %s" % (bitmap_path))
			new_img.set_bitmap(bitmap_path)
			#new_img.chromakey_colour = (0, 255, 0)
			#new_img.chromakey_colour = None
			new_img.position = (self.screen_height / 2, self.screen_width / 2)
			new_img.set_size((self.screen_height, self.screen_width))
			self.current_queue.append(new_img)
		
	def get_next_image(self):
		#print("Debug: queue length is %s" % (len(self.current_queue)))
		if len(self.current_queue) > 0:
			self.current_image = self.current_queue[0]
			#print("Debug: new image path is '%s'" % (self.current_image.bitmap_file_path))
			new_queue = []
			if len(self.current_queue) > 1:
				for i in range(1, len(self.current_queue)):
					#print("Debug: added one image from queue")
					new_queue.append(self.current_queue[i])
			self.current_queue = new_queue
		else:
			#print("Debug: no more images in queue - resetting")
			self.reset()
			#self.get_next_image()
					
	def main_loop(self):
		elapsed_time = time.time() - self.image_start_time
		#print("Debug: elapsed time for current image is '%s'" % (elapsed_time))
		#print("Debug: self.elapsed_millisecond_counter is '%s'" % (self.elapsed_millisecond_counter))
		if elapsed_time >= self.seconds_per_image:
			#print("Debug: current image has expired")
			self.image_start_time = time.time()
			self.get_next_image()
		if self.current_image:
			self.on_screen_objects = [ self.current_image ]
		else:
			self.on_screen_objects = []
			self.get_next_image()
			#self.on_screen_objects = [ self.current_image ]

class concentric_rings(object_manager):
	def __init__(self, screen, screen_width, screen_height):
		super(concentric_rings, self).__init__(screen, screen_width, screen_height)
		self.min_center_gap = 0.0
		self.max_center_gap = 100.0
		self.min_ring_thickness = 10.0
		self.max_ring_thickness = 100.0
		self.min_ring_spacing = 2.0
		self.max_ring_spacing = 50.0
		self.screen = screen
		self.screen_width = screen_width
		self.screen_height = screen_height
		self.min_initial_velocity = -5.0
		self.max_initial_velocity = 5.0
		self.min_absolute_velocity = 0.2
		self.min_segment_length_in_degrees = 5.0
		self.max_segment_length_in_degrees = 320.0
		self.min_segment_spacing_in_degrees = 5.0
		self.max_segment_spacing_in_degrees = 320.0
		
		self.colour_background = (0, 0, 0)
		self.colour_foreground = (230, 255, 245)
		self.reset_on_no_drawing = False
		self.reset_on_no_objects = False
		self.colours = self.get_spooky_colours()
		
	def reset(self):
		super(slow_polygram_rotating_coloured, self).reset()
		
		for i in range(0, len(self.on_screen_objects)):
			self.on_screen_objects[i].colour_foreground = new_colour
			
	def reset(self):
		self.current_settings_start_time = time.time()
		colour_num = random.randint(0, (len(self.colours) - 1))
		new_colour = self.colours[colour_num]
		self.colour_foreground = new_colour
		center_gap = random.uniform(self.min_center_gap, self.max_center_gap)
		ring_thickness = random.uniform(self.min_ring_thickness, self.max_ring_thickness)
		ring_spacing = random.uniform(self.min_ring_spacing, self.max_ring_spacing)
		longer_edge = self.screen_width
		if self.screen_width < self.screen_height:
			longer_edge = self.screen_height
		longer_edge *= 1.1
		num_rings = int(math.ceil(((longer_edge / 2) - center_gap) / (ring_thickness + ring_spacing)))

		self.on_screen_objects = []
		#new_ring = segmented_ring(200.0, 100.0)
		#new_ring.position = (self.screen_width / 2.0, self.screen_height / 2.0)
		#new_ring.colour = self.colour_foreground
		#new_ring.angular_velocity = 1.0
		#new_ring.velocity = (0.0, 0.0)
		#self.on_screen_objects.append(new_ring)
		current_ring_radius = center_gap + ring_thickness
		#if current_ring_radius < ring_thickness:
		#	current_ring_radius = ring_thickness
		for i in range(0, num_rings):
			#new_ring = segmented_ring(center_gap + (i * (ring_thickness + ring_spacing)), ring_thickness)
			new_ring = segmented_ring(current_ring_radius, ring_thickness)
			current_ring_radius += (ring_thickness + ring_spacing)
			new_ring.position = (self.screen_width / 2.0, self.screen_height / 2.0)
			new_ring.colour = self.colour_foreground
			new_ring.angular_velocity = random.uniform(self.min_initial_velocity, self.max_initial_velocity)
			new_ring.colour = self.colour_foreground
			while abs(new_ring.angular_velocity) < self.min_absolute_velocity:
				new_ring.angular_velocity = random.uniform(self.min_initial_velocity, self.max_initial_velocity)
			
			current_degree = 0.0
			while current_degree < 360.0:
				segment_start = current_degree
				segment_end = segment_start + random.uniform(self.min_segment_length_in_degrees, self.max_segment_length_in_degrees)
				segment_spacing = random.uniform(self.min_segment_spacing_in_degrees, self.max_segment_spacing_in_degrees)
				if (segment_end + segment_spacing) > 360.0:
					segment_end = 360.0 - segment_spacing
					current_degree = 360.0
				else:
					current_degree = segment_end + segment_spacing
				if segment_end > segment_start:
					rs = ring_segment()
					rs.start_degree = segment_start
					rs.end_degree = segment_end
					new_ring.segments.append(rs)
				else:
					current_degree = 360.0
		
			new_ring.velocity = (0.0, 0.0)
			self.on_screen_objects.append(new_ring)

if __name__=='__main__':
	script_name = sys.argv[0]
	
	base_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
	
	pygame.init()
	
	pygame.mouse.set_visible(0)
	
	screen_width = 800
	screen_height = 480

	terminal_lines = ""
	f = open(os.path.join(base_path, 'terminal_text_cyberpunk.txt'), 'r')
	terminal_lines = f.read()
	f.close()
		
	slideshow_images = os.path.join(base_path, 'images')
	
	slideshow_image_file_list_base = [f for f in os.listdir(slideshow_images) if os.path.isfile(os.path.join(slideshow_images, f))]
	
	slideshow_image_file_list = []
	
	for sif in slideshow_image_file_list_base:
		slideshow_file = os.path.join(slideshow_images, sif)
		if "." in sif:
			slideshow_file_extension = sif.lower()
			last_dot_pos = slideshow_file_extension.rfind('.')
			slideshow_file_extension = slideshow_file_extension[last_dot_pos + 1:]
			#print(slideshow_file_extension)
			if slideshow_file_extension in [ 'bmp', 'gif', 'jpg', 'jpeg', 'png', 'tif', 'tiff', 'tga' ]:
				slideshow_image_file_list.append(slideshow_file)
	
	#print(slideshow_image_file_list)
	
	#sys.exit(0)

	clock = pygame.time.Clock()
	
	screen = pygame.display.set_mode((screen_width, screen_height), pygame.FULLSCREEN)

	blurred_width = screen_width / 4
	blurred_height = screen_height / 4
	
	#colour_background = (255, 255, 255)
	#colour_foreground = (0, 0, 0)

	colour_background = (0, 0, 0)
	# 80s vector display white
	colour_foreground = (230, 255, 245)
	# ancient amber monitor
	#colour_foreground = (255, 194, 130)
	# Alien 3 computer green
	#colour_foreground = (27, 255, 90)
	# ominous red
	#colour_foreground = (200, 0, 0)
	colour_foreground_alt = (255, 0, 255)
	blurred_width = screen_width / 4
	blurred_height = screen_height / 4
	blit_instead_of_draw = True
	chromakey_colour = colour_background
	# 144 or 128 is probably best for this on a desktop display
	# 144 is fine, but maybe 192 or 224 for the Raspberry Pi 7" touchscreen
	global_alpha = 128
	global_alpha = 144
	# 4 for a desktop display, 16 or even 32 for the Raspberry Pi 7" touchscreen
	#screen_fade_alpha = 4
	screen_fade_alpha = 16
	
	target_fps = 60
	
	shuffle_mode = True
	shuffle_mode_interval_seconds = 31.337
	#shuffle_mode_interval_seconds = 1.337

	
	s1 = pygame.Surface((screen_width, screen_height))
	s2 = pygame.Surface((screen_width, screen_height))
	s3 = pygame.Surface((screen_width, screen_height))
	s4 = pygame.Surface((screen_width, screen_height))
	#s2_blurred = pygame.Surface((screen_width, screen_height))
	s1.set_alpha(screen_fade_alpha)                # alpha level
	s2.set_alpha(255)                # alpha level
	#s2_blurred.set_alpha(200)
	s3.set_alpha(255)                # alpha level	
	s3.set_colorkey(colour_background)
	screen.fill(colour_background)
	s1.fill(colour_background)
	
	running = True
	
	add_blur = True
	#add_blur = False
	obj_managers = []

	mgr_rings = concentric_rings(screen, screen_width, screen_height)
	mgr_rings.reset()
	obj_managers.append(mgr_rings)
	
	#mgr_slow_polygon_rotating_white = slow_polygon_rotating_white(screen, screen_width, screen_height)
	#mgr_slow_polygon_rotating_white.reset()
	#obj_managers.append(mgr_slow_polygon_rotating_white)
	
	mgr_matrix_code = matrix_code(screen, screen_width, screen_height)
	mgr_matrix_code.reset()
	#mgr_matrix_code.set_font('EversonMonoBold.ttf', 24)
	mgr_matrix_code.set_font('mplus-1m-medium.ttf', 24)
	obj_managers.append(mgr_matrix_code)
	
	mgr_hexgrid_2 = hexgrid_2(screen, screen_width, screen_height)
	mgr_hexgrid_2.reset()
	obj_managers.append(mgr_hexgrid_2)
	
	mgr_shuffled_slideshow = image_slideshow(screen, screen_width, screen_height)
	# mgr_shuffled_slideshow.image_paths.append('images/placeholder-01.png')
	# mgr_shuffled_slideshow.image_paths.append('images/placeholder-02.png')
	# mgr_shuffled_slideshow.image_paths.append('images/placeholder-03.png')
	# mgr_shuffled_slideshow.image_paths.append('images/placeholder-04.png')
	# mgr_shuffled_slideshow.image_paths.append('images/placeholder-05.png')
		
	# mgr_shuffled_slideshow.image_paths.append('images/Soul_Reaver_Prototypes.png')
	# mgr_shuffled_slideshow.image_paths.append('images/Cindergarden.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/TBM1.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/TBM2.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/Kali.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/Soul_Reaver_Prototypes.png')
	# mgr_shuffled_slideshow.image_paths.append('images/Altered_Carbon.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/Bacteria.PNG')
	# mgr_shuffled_slideshow.image_paths.append('images/Cyberdyne.png')
	# mgr_shuffled_slideshow.image_paths.append('images/Rekall.png')
	# mgr_shuffled_slideshow.image_paths.append('images/Weyland-Yutani.png')
	#mgr_shuffled_slideshow.reset()

	for slideshow_path in slideshow_image_file_list:
		mgr_shuffled_slideshow.image_paths.append(slideshow_path)
	obj_managers.append(mgr_shuffled_slideshow)
	
	mgr_terminal_text_output_cyberpunk = terminal_text_infinite_loop(screen, screen_width, screen_height)
	mgr_terminal_text_output_cyberpunk.reset()
	#mgr_terminal_text_output_cyberpunk.set_font('NotoMono-Regular.ttf', 12)
	mgr_terminal_text_output_cyberpunk.set_font('NotoMono-Regular.ttf', 24)
	mgr_terminal_text_output_cyberpunk.characters_per_loop = 1.0
	#mgr_terminal_text_output_cyberpunk.characters_per_loop = 3.0
	#mgr_terminal_text_output_cyberpunk.add_text_to_buffer(terminal_lines)
	mgr_terminal_text_output_cyberpunk.base_text = terminal_lines
	obj_managers.append(mgr_terminal_text_output_cyberpunk)
	
	mgr_sweeping_lines_no_bounce_rotating = sweeping_lines_no_bounce(screen, screen_width, screen_height)
	mgr_sweeping_lines_no_bounce_rotating.enable_rotation = True
	mgr_sweeping_lines_no_bounce_rotating.reset()
	obj_managers.append(mgr_sweeping_lines_no_bounce_rotating)
	
	#mgr_sweeping_lines_no_bounce = sweeping_lines_no_bounce(screen, screen_width, screen_height)
	#mgr_sweeping_lines_no_bounce.reset()
	#obj_managers.append(mgr_sweeping_lines_no_bounce)
	
	#mgr_hexgrid_1 = hexgrid_1(screen, screen_width, screen_height)
	#mgr_hexgrid_1.reset()
	#obj_managers.append(mgr_hexgrid_1)
	
	#mgr_sweeping_lines = sweeping_lines(screen, screen_width, screen_height)
	#mgr_sweeping_lines.reset()
	#obj_managers.append(mgr_sweeping_lines)

	mgr_slow_polygram_rotating_coloured = slow_polygram_rotating_coloured(screen, screen_width, screen_height)
	mgr_slow_polygram_rotating_coloured.reset()
	obj_managers.append(mgr_slow_polygram_rotating_coloured)
	
	mgr_slow_polygrams_rotating_coloured = slow_polygrams_rotating_coloured(screen, screen_width, screen_height)
	mgr_slow_polygrams_rotating_coloured.reset()
	obj_managers.append(mgr_slow_polygrams_rotating_coloured)
	
	mgr_soul_reaver_prototypes = image_slideshow(screen, screen_width, screen_height)
	mgr_soul_reaver_prototypes.image_paths.append('images/Soul_Reaver_Prototypes.png')
	obj_managers.append(mgr_soul_reaver_prototypes)
	
	mgr_fast_polygrams_rotating_coloured = fast_polygrams_rotating_coloured(screen, screen_width, screen_height)
	mgr_fast_polygrams_rotating_coloured.reset()
	obj_managers.append(mgr_fast_polygrams_rotating_coloured)
	
	#mgr_slow_polygram_rotating_white = slow_polygram_rotating_white(screen, screen_width, screen_height)
	#mgr_slow_polygram_rotating_white.reset()
	#obj_managers.append(mgr_slow_polygram_rotating_white)
	
	#mgr_slow_polygon_rotating_white = slow_polygon_rotating_white(screen, screen_width, screen_height)
	#mgr_slow_polygon_rotating_white.reset()
	#obj_managers.append(mgr_slow_polygon_rotating_white)

	#mgr_fast_hex_white = fast_hex_white(screen, screen_width, screen_height)
	#mgr_fast_hex_white.reset()
	#obj_managers.append(mgr_fast_hex_white)

	#mgr_slow_polygon_white = slow_polygon_white(screen, screen_width, screen_height)
	#mgr_slow_polygon_white.reset()
	#obj_managers.append(mgr_slow_polygon_white)

	mgr_fast_polygon_white = fast_polygon_white(screen, screen_width, screen_height)
	mgr_fast_polygon_white.reset()
	obj_managers.append(mgr_fast_polygon_white)

	mgr_default = object_manager(screen, screen_width, screen_height)
	mgr_default.reset()
	obj_managers.append(mgr_default)

	active_manager_num = 0
	
	milliseconds_elapsed_last_tick = None
	
	start_time_shuffle_iteration = time.time()
	
	capture_screenshots = False
	screenshot_base_date = datetime.datetime.now()
	screenshot_number = 1
	
	while running:
		active_manager = obj_managers[active_manager_num]
		
		if shuffle_mode:
			elapsed = time.time() - start_time_shuffle_iteration
			#if elapsed >= shuffle_mode_interval_seconds:
			if elapsed >= active_manager.display_seconds:
				start_time_shuffle_iteration = time.time()
				active_manager_num += 1
				if active_manager_num >= len(obj_managers):
					active_manager_num = 0

		#s.set_alpha(128)
		screen.fill(colour_background)
		#s1.fill(colour_background)           # this fills the entire surface

		s2.blit(s1, (0,0))		
		
		if active_manager.enable_chromakey:
			s4.fill(chromakey_colour)
		else:
			s4.fill(colour_background)
		s4.set_alpha(global_alpha)
		if active_manager.enable_chromakey:
			s4.set_colorkey(chromakey_colour)
		else:
			s4.set_colorkey(None)
		#print("%s objects on-screen" % (len(active_manager.on_screen_objects)))
		if active_manager.reset_on_no_objects:
			if len(active_manager.on_screen_objects) < 1:
				active_manager.reset()
		active_manager.main_loop()			
		# draw is split out separately into its own loop in case objects are removed during main_loop
		#for i in range(0, len(active_manager.on_screen_objects)):
		#	active_manager.on_screen_objects[i].draw(s4)
		s4 = active_manager.draw(s4)
			
		s2.blit(s4, (0,0))
			
				
		if add_blur and active_manager.add_post_blur:
			#s2_blurred = pygame.transform.scale(pygame.transform.smoothscale(s2, (blurred_width, blurred_height)), (screen_width, screen_height))
			#s3.blit(s2, (0, 0))
			s2_blurred = pygame.transform.smoothscale(pygame.transform.smoothscale(s2, (blurred_width, blurred_height)), (screen_width, screen_height))
			s2_blurred.set_colorkey(colour_background)
			s2_blurred.set_alpha(128)
			s3.blit(s2, (0, 0))
			s3.blit(s2_blurred, (0, 0))
			screen.blit(s3, (0,0))
			#screen.blit(s2_blurred, (0,0))
		else:
			screen.blit(s2, (0,0))

		# update the screen
		pygame.display.update()
		
		if capture_screenshots:
			datestring = screenshot_base_date.strftime("%Y-%m-%d-%H-%M-%S")
			filename = "screenshots/screenshot-%s-%08d.png" % (datestring, screenshot_number)
			screenshot_number += 1
			pygame.image.save(screen, filename)
			
			
		milliseconds_elapsed_last_tick = clock.tick(target_fps)
		
		# check for keyboard events
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				pygame.quit(); sys.exit();
			elif event.type == pygame.KEYDOWN:
				if event.key == pygame.K_ESCAPE or event.key == pygame.K_q or event.key == pygame.K_x or event.key == pygame.K_KP_MULTIPLY:
					running = False
				if event.key == pygame.K_r or event.key == pygame.K_KP_ENTER:
					#for i in range(0, len(on_screen_objects)):
					#	on_screen_objects[i].position = (screen_width / 2, screen_height / 2)
					active_manager.reset()
					
				if event.key == pygame.K_s:
					if capture_screenshots:
						capture_screenshots = False
					else:
						capture_screenshots = True

				if event.key == pygame.K_0 or event.key == pygame.K_KP0:
					shuffle_mode = True
					
				if event.key == pygame.K_KP_MINUS or event.key == pygame.K_MINUS:
					shuffle_mode = False
					active_manager_num -= 1
					if active_manager_num < 0:
						active_manager_num = len(obj_managers) - 1
				if event.key == pygame.K_KP_PLUS or event.key == pygame.K_PLUS:
					shuffle_mode = False
					active_manager_num += 1
					if active_manager_num >= len(obj_managers):
						active_manager_num = 0

				switch_to_manager_index = 0
				if event.key == pygame.K_1 or event.key == pygame.K_KP1:
					switch_to_manager_index = 1
				if event.key == pygame.K_2 or event.key == pygame.K_KP2:
					switch_to_manager_index = 2
				if event.key == pygame.K_3 or event.key == pygame.K_KP3:
					switch_to_manager_index = 3
				if event.key == pygame.K_4 or event.key == pygame.K_KP4:
					switch_to_manager_index = 4
				if event.key == pygame.K_5 or event.key == pygame.K_KP5:
					switch_to_manager_index = 5
				if event.key == pygame.K_6 or event.key == pygame.K_KP6:
					switch_to_manager_index = 6
				if event.key == pygame.K_7 or event.key == pygame.K_KP7:
					switch_to_manager_index = 7
				if event.key == pygame.K_8 or event.key == pygame.K_KP8:
					switch_to_manager_index = 8
				if event.key == pygame.K_9 or event.key == pygame.K_KP9:
					switch_to_manager_index = 9
					
				if switch_to_manager_index > 0:
					shuffle_mode = False
					new_manager_index = switch_to_manager_index - 1 % (len(obj_managers) - 1)
					active_manager_num = new_manager_index
		
		if not running:
			#break
			pygame.mouse.set_visible(1)
			sys.exit(0)
	
	pygame.mouse.set_visible(1)
	
	
	# screenshot-2019-08-01-22-41-30-00029099.png and previous go in the sweeping lines folder